#include "stdafx.h"
#include <afx.h>
#include <afxtempl.h>
#include "complex.h"
#include "problem.h"
#include "femmview.h"
#include "xyplot.h"
#include "MainFrm.h"
#include "femmviewDoc.h"
	
#define uo (PI*4.e-07)

BOOL CFemmviewDoc::ContourClosed()
{
	int n=contour.GetSize()-1;
	if (n<=1) return FALSE;
	if (abs(contour[0]-contour[n])<1.e-06) return TRUE;
	return FALSE;
}

BOOL CFemmviewDoc::SlowInContour(double x, double y)
{
	// InContour test that uses an angle sum.  Doesn't have any special
	// cases, but it's slooow because it repeatedly evaluates arctan.
	int i;
	CComplex p;
	double z;

	for(i=0,z=0,p.Set(x,y);i<contour.GetSize()-1;i++)
		z+=arg((contour[i+1]-p)/(contour[i]-p));

	if (fabs(z)<0.01) return FALSE;
 	else return TRUE;
}

BOOL CFemmviewDoc::InContour(double x, double y)
{	
	if (ContourClosed()==FALSE) return FALSE;

	int i,k;
	CComplex p0,p1,p2;
	BOOL flag;
	double z,xo;

	// use ray crossing test. Ray goes to the right of the test point;
	for(i=0,k=-1;i<contour.GetSize()-1;i++)
	{
		flag=TRUE;
		p0=contour[i]; p1=contour[i+1];
		// if both points are to the left, there is no crossing...
		if ((p0.re<=x) && (p1.re<=x)) flag=FALSE;
		// if both points are above, there is no crossing...
		if ((p0.im>y) && (p1.im>y)) flag=FALSE;
		// if both points are below, there is no crossing...
		if ((p0.im<y) && (p1.im<y)) flag=FALSE;
		// if both points are on the same y, can't really test for crossing;
		if (p0.im==p1.im) flag=FALSE;

		if (flag==TRUE){
			if ((y==p0.im) || (y==p1.im))
				return SlowInContour(x,y);
			z=(y-p0.im)/(p1.im-p0.im);
			xo=p0.re + z*(p1.re - p0.re);
			if (xo<=x) flag=FALSE;
		}

		if (flag==TRUE) k*=(-1);
	}

	if (k>0) return TRUE;

	return FALSE;
}

BOOL CFemmviewDoc::InContour(CElement &elm)
{
	return InContour(elm.ctr.re,elm.ctr.im);
}

BOOL CFemmviewDoc::InContour(CComplex p)
{
	return InContour(p.re,p.im);
}

BOOL CFemmviewDoc::OnContour(CElement &elm)
{
	int i,k;

	for(i=0,k=0;i<3;i++)
		k+=InContour(meshnode[elm.p[i]].CC());

	if ((k>0) && (k<3)) return TRUE;
	return FALSE;
}

void CFemmviewDoc::AvgMaxwell(double &fx, double &fy)
{
	// special version of Maxwell's Stress Tensor
	// integration that takes integrations over entire
	// elements, rather than on lines that cut through
	// the elements.  This approach seems to be more
	// accurate than the "regular" stress tensor.
	int *u;
	int i,j,k;
	int n[3];
	double b0,b1,c0,c1,d,R;
	double nx,ny,bx,by;
	double bdotn,bdotb;
	double U[3];
	
	u=(int *)calloc(meshnode.GetSize(),sizeof(int));
	
	for(i=0;i<meshnode.GetSize();i++) 
		u[i]=1-InContour(meshnode[i].CC());
	
	for(i=0,fx=0.,fy=0.;i<meshelem.GetSize();i++)
	{
		for(j=0,k=0;j<3;j++){
			n[j]=meshelem[i].p[j];
			U[j]=(double) u[n[j]];
			k+=u[n[j]];
		}
		if ((k==1) || (k==2))
		{
			b0=meshnode[n[2]].y - meshnode[n[0]].y;
			b1=meshnode[n[0]].y - meshnode[n[1]].y;
			c0=meshnode[n[0]].x - meshnode[n[2]].x;
			c1=meshnode[n[1]].x - meshnode[n[0]].x;
			for(k=0,R=0.;k<3;k++) R+=meshnode[n[k]].x;
			b0*=LengthConv[LengthUnits];
			b1*=LengthConv[LengthUnits];
			c0*=LengthConv[LengthUnits];
			c1*=LengthConv[LengthUnits];
			R *=LengthConv[LengthUnits]/3.;
			d=(b0*c1-b1*c0);

			nx=((U[1]-U[0])*b0 + (U[2]-U[0])*b1)/d;
			ny=((U[1]-U[0])*c0 + (U[2]-U[0])*c1)/d;

			bx=meshelem[i].B1.re;
			by=meshelem[i].B2.re;

			bdotn = bx*nx + by*ny;
			bdotb = bx*bx + by*by;

			if(ProblemType==0)
			{
				fx += d*(bx*bdotn - 0.5*nx*bdotb)/(2.*uo);
				fy += d*(by*bdotn - 0.5*ny*bdotb)/(2.*uo);
			}
			else{
				fy += PI*R*d*(by*bdotn - 0.5*ny*bdotb)/uo;
			}
		}
	}

	free(u);
}

void CFemmviewDoc::AvgMaxwell(double &fx,  double &fy,
							  CComplex &f2x, CComplex &f2y)
{
	// special version of Maxwell's Stress Tensor
	// integration that takes integrations over entire
	// elements, rather than on lines that cut through
	// the elements.  This approach seems to be more
	// accurate than the "regular" stress tensor.
	// This is the version for harmonic problems.

	int *u;
	int i,j,k;
	int n[3];
	double b0,b1,c0,c1,d,R;
	double nx,ny;
	double Bxr,Bxi,Byr,Byi;
	double U[3];
	
	u=(int *)calloc(meshnode.GetSize(),sizeof(int));
	
	for(i=0;i<meshnode.GetSize();i++) 
		u[i]=1-InContour(meshnode[i].CC());
	
	for(i=0,fx=0.,fy=0.,f2x=0.,f2y=0.;i<meshelem.GetSize();i++)
	{
		for(j=0,k=0;j<3;j++){
			n[j]=meshelem[i].p[j];
			U[j]=(double) u[n[j]];
			k+=u[n[j]];
		}
		if ((k==1) || (k==2))
		{
			b0=meshnode[n[2]].y - meshnode[n[0]].y;
			b1=meshnode[n[0]].y - meshnode[n[1]].y;
			c0=meshnode[n[0]].x - meshnode[n[2]].x;
			c1=meshnode[n[1]].x - meshnode[n[0]].x;
			for(k=0,R=0.;k<3;k++) R+=meshnode[n[k]].x;
			b0*=LengthConv[LengthUnits];
			b1*=LengthConv[LengthUnits];
			c0*=LengthConv[LengthUnits];
			c1*=LengthConv[LengthUnits];
			R *=LengthConv[LengthUnits]/3.;
			d=(b0*c1-b1*c0);

			nx=((U[1]-U[0])*b0 + (U[2]-U[0])*b1)/d;
			ny=((U[1]-U[0])*c0 + (U[2]-U[0])*c1)/d;

			Bxr=meshelem[i].B1.re;
			Bxi=meshelem[i].B1.im;
			Byr=meshelem[i].B2.re;
			Byi=meshelem[i].B2.im;

			// All of this mess comes out of Mathematica.
			// I explicitly substituted flux as a function of
			// time into the stress tensor and then sorted out
			// the steady-state and 2x components.  This isn't
			// pretty, but it was the easiest way to make sure
			// not to screw up.
			if(ProblemType==0)
			{
				fx += (d/(2.*uo))*((Bxi*Bxi*nx + Bxr*Bxr*nx - 
					  (Byi*Byi + Byr*Byr)*nx + 2.*Bxi*Byi*ny + 
						 2.*Bxr*Byr*ny)/4.);
				fy += (d/(2.*uo))*((2.*Bxi*Byi*nx + 
					2.*Bxr*Byr*nx - Bxi*Bxi*ny - 
					Bxr*Bxr*ny + (Byi*Byi + Byr*Byr)*ny)/4.);
				f2x += (d/(2.*uo))*(-(((Bxi - I*Bxr)*(Bxi - I*Bxr)
						- (Byi - I*Byr)*(Byi - I*Byr))*nx) - 
						2.*(Bxi - I*Bxr)*(Byi - I*Byr)*ny)/4. ;
				f2y += (d/(2.*uo))*((-2.*(Bxi - I*Bxr)*(Byi - I*Byr)*nx + 
					   ((Bxi - I*Bxr)*(Bxi - I*Bxr) - 
					   (Byi - I*Byr)*(Byi - I*Byr))*ny)/4.);
			}
			else{
				fy += (PI*R*d/uo)*((2.*Bxi*Byi*nx + 
					2.*Bxr*Byr*nx - Bxi*Bxi*ny - 
					Bxr*Bxr*ny + (Byi*Byi + Byr*Byr)*ny)/4.);
				f2y += (PI*R*d/uo)*((-2.*(Bxi - I*Bxr)*(Byi - I*Byr)*nx + 
					   ((Bxi - I*Bxr)*(Bxi - I*Bxr) - 
					   (Byi - I*Byr)*(Byi - I*Byr))*ny)/4.);
			}
		}
	}

	free(u);
}
