#include<stdafx.h>
#include "complex.h"
#include<stdio.h>

CComplex::CComplex()
{
	re=0.;	im=0.;
}

CComplex::CComplex(double x, double y)
{
	re=x; im=y;
}

CComplex CComplex::Sqrt()
{
	double w,z;
	CComplex y;

	if((re==0) && (im==0)) w=0;
	else if(fabs(re)>fabs(im)){
		z=im/re;
		w=sqrt(fabs(re))*sqrt( (1.+sqrt(1.+z*z))/2. );
	}
	else{
		z=re/im;
		w=sqrt(fabs(im))*sqrt( (fabs(z)+sqrt(1.+z*z))/2. );
	}

	if(w==0){
		y.re=0;
		y.im=0;
		return y;
	}

	if(re>=0){
		y.re=w;
		y.im=im/(2.*w);
		return y;
	}

	if(im>=0){
		y.re=fabs(im)/(2.*w);
		y.im=w;
		return y;
	}

	y.re=fabs(im)/(2.*w);
	y.im= (-w);
	return y;
}

CComplex CComplex::Conj()
{
	return CComplex(re,-im);
}

double CComplex::Abs()
{
	if ((re==0) && (im==0)) return 0.;

	if (fabs(re)>fabs(im))
		return fabs(re)*sqrt(1.+(im/re)*(im/re));
	else
		return fabs(im)*sqrt(1.+(re/im)*(re/im));
}

double CComplex::Arg()
{
	if ((re==0) && (im==0)) return 0.;

	return atan2(im,re);
}

CComplex CComplex::Inv()
{
	double c;
	CComplex z;

	if(fabs(re)>fabs(im))
	{
		c=im/re;
		z.re=1./(re*(1.+c*c));
		z.im=(-c)*z.re;
	}
	else{
		c=re/im;
		z.im=(-1.)/(im*(1.+c*c));
		z.re=(-c)*z.im;
	}

	return z;
}

double CComplex::Re()
{
	return re;
}

double CComplex::Im()
{
	return im;
}

void CComplex::Set(double x, double y)
{
	re=x; im=y;
}

char* CComplex::ToString(char *s)
{
	if (im<0)
		sprintf(s,"%.3e - j %.3e",re,fabs(im));
	else
		sprintf(s,"%.3e + j %.3e",re,im);
	return s;
}

//******* Addition ***************************************************

CComplex CComplex::operator+( CComplex z )
{
	return CComplex(re+z.re,im+z.im);
};

CComplex CComplex::operator+( int z )
{
	return CComplex(re+((double) z),im);
};

CComplex CComplex::operator+( double z )
{
	return CComplex(re+z,im);
};

CComplex CComplex::operator+=( CComplex z )
{
	re+=z.re;
	im+=z.im;
	return CComplex(re,im);
};

CComplex CComplex::operator+=( double z )
{
	re+=z;
	return CComplex(re,im);
};

CComplex CComplex::operator+=( int z )
{
	re+=(double) z;
	return CComplex(re,im);
};

CComplex operator+( int x, CComplex y )
{
	return CComplex( ((double) x) + y.re, y.im );
}

CComplex operator+( double x, CComplex y )
{
	return CComplex( x + y.re, y.im );
}



//******* Subtraction ***************************************************
CComplex CComplex::operator-()
{
	return CComplex(-re,-im);
}

CComplex CComplex::operator-( CComplex z )
{
	return CComplex(re-z.re,im-z.im);
};

CComplex CComplex::operator-( int z )
{
	return CComplex(re-((double) z),im);
};

CComplex CComplex::operator-( double z )
{
	return CComplex(re-z,im);
};

CComplex CComplex::operator-=( CComplex z )
{
	re-=z.re;
	im-=z.im;
	return CComplex(re,im);
};

CComplex CComplex::operator-=( double z )
{
	re-=z;
	return CComplex(re,im);
};

CComplex CComplex::operator-=( int z )
{
	re-=(double) z;
	return CComplex(re,im);
};

CComplex operator-( int x, CComplex y )
{
	return CComplex( ((double) x) - y.re, - y.im );
}

CComplex operator-( double x, CComplex y )
{
	return CComplex( x - y.re, - y.im );
}

//******* Multiplication ***************************************************

CComplex CComplex::operator*( CComplex z )
{
	return CComplex(re*z.re - im*z.im,re*z.im + im*z.re);
};

CComplex CComplex::operator*( int z )
{
	return CComplex( re*((double) z),im*((double) z) );
};

CComplex CComplex::operator*( double z )
{
	return CComplex(re*z,im*z);
};

CComplex CComplex::operator*=( CComplex z )
{
	CComplex x(re*z.re - im*z.im,re*z.im + im*z.re);
	re=x.re; im=x.im;
	return x;
};

CComplex CComplex::operator*=( double z )
{
	re*=z; im*=z;
	return CComplex(re,im);
};

CComplex CComplex::operator*=( int z )
{
	re*=(double) z;
	im*=(double) z;
	return CComplex(re,im);
};

CComplex operator*( int x, CComplex y )
{
	return CComplex( ((double) x) * y.re, ((double) x)*y.im );
}

CComplex operator*( double x, CComplex y )
{
	return CComplex( x*y.re, x*y.im );
}

//******* Division ***************************************************

CComplex CComplex::operator/( CComplex z )
{
	return CComplex(re,im)*z.Inv();
};


CComplex CComplex::operator/( int z )
{
	return CComplex(re/((double) z),im/((double) z));
};

CComplex CComplex::operator/( double z )
{
	return CComplex(re/z,im/z);
};

CComplex CComplex::operator/=( CComplex z )
{
	CComplex x(re,im);
	x=x/z;
	re=x.re;
	im=x.im;

	return x;
};

CComplex CComplex::operator/=( double z )
{
	re/=z;
	im/=z;
	return CComplex(re,im);
};

CComplex CComplex::operator/=( int z )
{
	re/=(double) z;
	im/=(double) z;
	return CComplex(re,im);
};

CComplex operator/( int x, CComplex y )
{
	return x*y.Inv();
}

CComplex operator/( double x, CComplex y )
{
	return x*y.Inv();
}

//****** Equals definitions ********************************

CComplex CComplex::operator=(double z)
{
	re=z;
	im=0;

	return CComplex(re,im);
}

CComplex CComplex::operator=(int z)
{
	re=(double) z;
	im=0;

	return CComplex(re,im);
}

//***** Tests ***********************************************
BOOL CComplex::operator==(CComplex z){
	if ((z.im==im) && (z.re==re)) return TRUE;
	return FALSE;
}

BOOL CComplex::operator==(double z){
	if ((z==re) && (im==0)) return TRUE;
	return FALSE;
}

BOOL CComplex::operator==(int z){
	if ((re==(double) z) && (im==0)) return TRUE;
	return FALSE;
}

BOOL CComplex::operator!=(CComplex z){
	if (z==CComplex(re,im)) return FALSE;
	return TRUE;
}

BOOL CComplex::operator!=(double z){
	if ((re!=z) || (im!=0)) return TRUE;
	return FALSE;
}

BOOL CComplex::operator!=(int z){
	if ((re!=(double) z) || (im!=0)) return TRUE;
	return FALSE;
}

//***** Useful functions ************************************
CComplex exp(CComplex x)
{
	CComplex y;
	
	y.re=cos(x.im)*exp(x.re);
	y.im=sin(x.im)*exp(x.re);
	
	return y;
}

CComplex sqrt(CComplex x)
{
	return x.Sqrt();
}

CComplex tanh(CComplex x)
{
	CComplex y;

	if (x.re>0){
		y=(1-exp(-2*x))/(1+exp(-2*x));
	}
	else{
		y=(exp(2*x)-1)/(exp(2*x)+1);
	}

	return y;
}

CComplex sinh(CComplex x)
{
	return (exp(x)-exp(-x))/2;
}

CComplex cosh(CComplex x)
{
	return (exp(x)+exp(-x))/2;
}

CComplex cos(CComplex x)
{
	return (exp(I*x)+exp(-I*x))/2;
}

CComplex acos(CComplex x)
{
	return PI/2. - arg(I*x + sqrt(1 - x*x)) + I*log(abs(I*x + sqrt(1 - x*x)));
}

CComplex sin(CComplex x)
{
	return (exp(I*x)-exp(-I*x))/(2*I);
}

CComplex asin(CComplex x)
{
	return arg(I*x + sqrt(1 - x*x)) - I*log(abs(I*x + sqrt(1 - x*x)));
}

CComplex tan(CComplex x)
{
	return sin(x)/cos(x);
}

CComplex atan(CComplex x)
{
	return (arg(1+I*x) - arg(1-I*x) - I*(log(abs(1+I*x)/abs(1-I*x))))/ 2;
}

double abs(CComplex x)
{
	return x.Abs();
}

double arg(CComplex x)
{
	return x.Arg();
}

CComplex log(CComplex x)
{
	CComplex y;

	y.im=x.Arg();
	y.re=log(x.Abs());

	return y;
}

CComplex pow(CComplex x, double y)
{
	return exp(y*log(x));
}

CComplex pow(CComplex x, int y)
{
	if (y==0) return CComplex(1,0);

	int i;
	CComplex z;

	if (y>0){
		z=x;
		for(i=1;i<y;i++) z=z*x;
	}
	else{
		z=x.Inv();
		for(i=1;i<(-y);i++) z=z*x.Inv();
	}

	return z;
}

CComplex pow(CComplex x, CComplex y)
{
	return exp(y*log(x));
}




// Support for old ways of dealing with complex operations...

CComplex Mult(CComplex a, CComplex b)
{
	return a*b;
}

CComplex Mult(double a, CComplex b)
{
	return a*b;
}

CComplex Mult(CComplex a, double b)
{
	return a*b;
}

CComplex Div(CComplex a, CComplex b)
{	
	return a/b;
}

CComplex Div(CComplex a, double b)
{	
	return a/b;
}

CComplex Div(double a, CComplex b)
{	
	return a/b;
}

CComplex Add(CComplex a, CComplex b)
{
	return a+b;
}

CComplex Add(double a, CComplex b)
{
	return a+b;
}

CComplex Add(CComplex b, double a)
{
	return a+b;
}

CComplex Subt(CComplex a, CComplex b)
{
	return a-b;
}

CComplex Subt(CComplex a, double b)
{
	return a-b;
}

CComplex Subt(double a, CComplex b)
{
	return a-b;
}

double cRe(CComplex a)
{
	return a.re;
}

double cIm(CComplex a)
{
	return a.im;
}

