// femmviewDoc.cpp : implementation of the CFemmviewDoc class
//


#include "stdafx.h"
#include <afx.h>
#include <afxtempl.h>
#include "complex.h"
#include "problem.h"
#include "femmview.h"
#include "xyplot.h"
#include "MainFrm.h"
#include "femmviewDoc.h"
#include "ptloc.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CFemmviewDoc

IMPLEMENT_DYNCREATE(CFemmviewDoc, CDocument)

BEGIN_MESSAGE_MAP(CFemmviewDoc, CDocument)
	//{{AFX_MSG_MAP(CFemmviewDoc)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CFemmviewDoc construction/destruction

double sqr(double x)
{
	return x*x;
}

CFemmviewDoc::CFemmviewDoc()
{
	// set some default values for problem definition
	Frequency=0.;
	LengthUnits=0;
	ProblemType=FALSE;
	ProblemNote="Add comments here.";
	FirstDraw=FALSE;
	A_High=0.;
	A_Low=0.;
	A_lb=0.;
	A_ub=0.;
	Smooth=TRUE;
	NumList=NULL;
	ConList=NULL;
	LengthConv=(double *)calloc(6,sizeof(double));
	LengthConv[0]=0.0254;	//inches
	LengthConv[1]=0.001;	//millimeters
	LengthConv[2]=0.01;		//centimeters
	LengthConv[3]=1.;		//meters
	LengthConv[4]=2.54e-05; //mils
	LengthConv[5]=1.e-06;	//micrometers
	
}

CFemmviewDoc::~CFemmviewDoc()
{
	int i;
	free(LengthConv);
	for(i=0;i<meshnode.GetSize();i++)
		if(ConList[i]!=NULL) free(ConList[i]);
	free(ConList);
	free(NumList);
}

BOOL CFemmviewDoc::OnNewDocument()
{
	if (!CDocument::OnNewDocument())
		return FALSE;

	// TODO: add reinitialization code here
	// (SDI documents will reuse this document)
	int i;

	// clear out all current lines, nodes, and block labels
	nodelist.RemoveAll();
	linelist.RemoveAll();
	blocklist.RemoveAll();
	arclist.RemoveAll();
	nodeproplist.RemoveAll();
	lineproplist.RemoveAll();
	blockproplist.RemoveAll();
	circproplist.RemoveAll();
	meshnode.RemoveAll();
	meshelem.RemoveAll();
	contour.RemoveAll();
	if(NumList!=NULL)
	{
		for(i=0;i<meshnode.GetSize();i++)
			if(ConList[i]!=NULL) free(ConList[i]);
		free(ConList); ConList=NULL;
		free(NumList); NumList=NULL;
	}

	// set problem attributes to generic ones;
	Frequency=0.;
	LengthUnits=0;
	ProblemType=FALSE;
	ProblemNote="Add comments here.";

	return TRUE;
}

/////////////////////////////////////////////////////////////////////////////
// CFemmviewDoc serialization

void CFemmviewDoc::Serialize(CArchive& ar)
{
	if (ar.IsStoring())
	{
		// TODO: add storing code here
	}
	else
	{
		// TODO: add loading code here
	}
}

/////////////////////////////////////////////////////////////////////////////
// CFemmviewDoc diagnostics

#ifdef _DEBUG
void CFemmviewDoc::AssertValid() const
{
	CDocument::AssertValid();
}

void CFemmviewDoc::Dump(CDumpContext& dc) const
{
	CDocument::Dump(dc);
}
#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CFemmviewDoc commands

char* StripKey(char *c)
{
	char *d;
	int i,k;

	k=strlen(c);

	for(i=0;i<k;i++){
		if (c[i] == '='){
			d=c+i+1;
			return d;
		}
	} 

	return c+k;
}

BOOL CFemmviewDoc::OnOpenDocument(LPCTSTR lpszPathName) 
{
	if (!CDocument::OnOpenDocument(lpszPathName))
		return FALSE;
	
	// TODO: Add your specialized creation code here
		
	FILE *fp;
	int i,j,k,t;
	char s[1024],q[1024];
	char *v;
	double x,b,bi,br;
	double zr,zi;
	BOOL flag=FALSE;
	CPointProp	  PProp;
	CBoundaryProp BProp;
	CMaterialProp MProp;
	CCircuit      CProp;
	CNode		node;
	CSegment	segm;
	CArcSegment asegm;
	CElement	elm;
	CBlockLabel blk;
	CMeshNode	mnode;
	CPoint		mline;

	// make sure old document is cleared out...
	if(NumList!=NULL){
		for(i=0;i<meshnode.GetSize();i++)
			if(ConList[i]!=NULL) free(ConList[i]);
		free(ConList); ConList=NULL;
		free(NumList); NumList=NULL;
	}
	nodelist.RemoveAll();		
	linelist.RemoveAll();
	blocklist.RemoveAll();
	arclist.RemoveAll();
	nodeproplist.RemoveAll();
	lineproplist.RemoveAll();
	blockproplist.RemoveAll();
	circproplist.RemoveAll();
	meshnode.RemoveAll();
	meshelem.RemoveAll();
	contour.RemoveAll();

	if ((fp=fopen(lpszPathName,"rt"))==NULL){
		fclose(fp);
		AfxMessageBox("Couldn't read from specified .poly file");
		return FALSE;
	}

	// define some defaults
	Frequency=0.;
	ProblemType=0;
	Coords=0;
	ProblemNote="";

	// parse the file
	while (flag==FALSE)
	{
		fgets(s,1024,fp);
		sscanf(s,"%s",q);

		// Deal with flag for file format version
		if( _strnicmp(q,"[format]",8)==0 ){
			float vers;
			v=StripKey(s);
			sscanf(v,"%f",&vers);
			if(vers>3.) AfxMessageBox("File is from a different version of femm\nTrying to read anyhow...");
			q[0]=NULL;
		}

		// Frequency of the problem
		if( _strnicmp(q,"[frequency]",11)==0){
			v=StripKey(s);
			sscanf(v,"%lf",&Frequency);
			q[0]=NULL;
		}

		// Units of length used by the problem
		if( _strnicmp(q,"[lengthunits]",13)==0){
			v=StripKey(s);
			sscanf(v,"%s",q);
			if( _strnicmp(q,"inches",6)==0) LengthUnits=0;
			else if( _strnicmp(q,"millimeters",11)==0) LengthUnits=1;
			else if( _strnicmp(q,"centimeters",1)==0) LengthUnits=2;
			else if( _strnicmp(q,"mils",4)==0) LengthUnits=4;
			else if( _strnicmp(q,"microns",6)==0) LengthUnits=5;
			else if( _strnicmp(q,"meters",6)==0) LengthUnits=3;
			q[0]=NULL;
		}

		// Problem Type (planar or axisymmetric)
		if( _strnicmp(q,"[problemtype]",13)==0){
			v=StripKey(s);
			sscanf(v,"%s",q);
			if( _strnicmp(q,"planar",6)==0) ProblemType=0;
			if( _strnicmp(q,"axisymmetric",3)==0) ProblemType=1;
			q[0]=NULL;
		}

		// Coordinates (cartesian or polar)
		if( _strnicmp(q,"[coordinates]",13)==0){
			v=StripKey(s);
			sscanf(v,"%s",q);
			if ( _strnicmp(q,"cartesian",4)==0) Coords=0;
			if ( _strnicmp(q,"polar",5)==0) Coords=1;
			q[0]=NULL;
		}
	
		// Comments
		if (_strnicmp(q,"[comment]",9)==0){
			v=StripKey(s);
			// put in carriage returns;
			k=strlen(v);
			for(i=0;i<k;i++)
				if((v[i]=='\\') && (v[i+1]=='n')){
					v[i]=13;
					v[i+1]=10;
				}

			for(i=0;i<k;i++)
				if(v[i]=='\"'){
					v=v+i+1;
					i=k;
				}
			k=strlen(v);
			if(k>0) for(i=k-1;i>=0;i--){
				if(v[i]=='\"'){
					v[i]=0;
					i=-1;
				}
			}
			ProblemNote=v;
			q[0]=NULL;
		}
		
		// Point Properties
		if( _strnicmp(q,"<beginpoint>",11)==0){	
			PProp.PointName="New Point Property";
			PProp.Jr=0.;
			PProp.Ji=0.;
			PProp.Ar=0.;
			PProp.Ai=0.;
			q[0]=NULL;
		}

		if( _strnicmp(q,"<pointname>",11)==0){
			v=StripKey(s);
			k=strlen(v);
			for(i=0;i<k;i++)
				if(v[i]=='\"'){
					v=v+i+1;
					i=k;
				}
			k=strlen(v);
			if(k>0) for(i=k-1;i>=0;i--){
				if(v[i]=='\"'){
					v[i]=0;
					i=-1;
				}
			}
			PProp.PointName=v;
			q[0]=NULL;
		}

		if( _strnicmp(q,"<A_re>",6)==0){
		   v=StripKey(s);
		   sscanf(v,"%lf",&PProp.Ar);
		   q[0]=NULL;
		}	

		if( _strnicmp(q,"<A_im>",6)==0){
		   v=StripKey(s);
		   sscanf(v,"%lf",&PProp.Ai);
		   q[0]=NULL;
		}
	
		if( _strnicmp(q,"<I_re>",6)==0){
		   v=StripKey(s);
		   sscanf(v,"%lf",&PProp.Jr);
		   q[0]=NULL;
		}	

		if( _strnicmp(q,"<I_im>",6)==0){
		   v=StripKey(s);
		   sscanf(v,"%lf",&PProp.Ji);
		   q[0]=NULL;
		}

		if( _strnicmp(q,"<endpoint>",9)==0){
			nodeproplist.Add(PProp);
			q[0]=NULL;
		}

		// Boundary Properties;
		if( _strnicmp(q,"<beginbdry>",11)==0){	
			BProp.BdryName="New Boundary";
			BProp.BdryFormat=0;
			BProp.A0=0.;
			BProp.A1=0.;
			BProp.A2=0.;
			BProp.phi=0.;	
			BProp.Mu=0.;
			BProp.Sig=0.;
			BProp.c0=0.;
			BProp.c1=0.;
			q[0]=NULL;
		}

		if( _strnicmp(q,"<bdryname>",10)==0){
			v=StripKey(s);
			k=strlen(v);
			for(i=0;i<k;i++)
				if(v[i]=='\"'){
					v=v+i+1;
					i=k;
				}
			k=strlen(v);
			if(k>0) for(i=k-1;i>=0;i--){
				if(v[i]=='\"'){
					v[i]=0;
					i=-1;
				}
			}
			BProp.BdryName=v;
			q[0]=NULL;
		}

		if( _strnicmp(q,"<bdrytype>",10)==0){
		   v=StripKey(s);
		   sscanf(v,"%i",&BProp.BdryFormat);
		   q[0]=NULL;
		}
		
		if( _strnicmp(q,"<mu_ssd>",8)==0){
		   v=StripKey(s);
		   sscanf(v,"%lf",&BProp.Mu);
		   q[0]=NULL;
		}	

		if( _strnicmp(q,"<sigma_ssd>",11)==0){
		   v=StripKey(s);
		   sscanf(v,"%lf",&BProp.Sig);
		   q[0]=NULL;
		}	

		if( _strnicmp(q,"<A_0>",5)==0){
		   v=StripKey(s);
		   sscanf(v,"%lf",&BProp.A0);
		   q[0]=NULL;
		}	

		if( _strnicmp(q,"<A_1>",5)==0){
		   v=StripKey(s);
		   sscanf(v,"%lf",&BProp.A1);
		   q[0]=NULL;
		}	
		
		if( _strnicmp(q,"<A_2>",5)==0){
		   v=StripKey(s);
		   sscanf(v,"%lf",&BProp.A2);
		   q[0]=NULL;
		}	

		if( _strnicmp(q,"<phi>",5)==0){
		   v=StripKey(s);
		   sscanf(v,"%lf",&BProp.phi);
		   q[0]=NULL;
		}	
		
		if( _strnicmp(q,"<c0>",4)==0){
		   v=StripKey(s);
		   sscanf(v,"%lf",&BProp.c0);
		   q[0]=NULL;
		}	
		
		if( _strnicmp(q,"<c1>",4)==0){
		   v=StripKey(s);
		   sscanf(v,"%lf",&BProp.c1);
		   q[0]=NULL;
		}	

		if( _strnicmp(q,"<endbdry>",9)==0){
			lineproplist.Add(BProp);
			q[0]=NULL;
		}


		// Block Properties;
		if( _strnicmp(q,"<beginblock>",12)==0){	
			MProp.BlockName="New Material";
			MProp.mu_x=1.;
			MProp.mu_y=1.;			// permeabilities, relative
			MProp.H_c=0.;				// magnetization, A/m
			MProp.Jr=0.;
			MProp.Ji=0.;				// applied current density, MA/m^2
			MProp.Cduct=0.;		    // conductivity of the material, MS/m
			MProp.Lam_d=0.;			// lamination thickness, mm
			MProp.Theta_h=0.;			// hysteresis angle, degrees
			MProp.LamFill=1.;			// lamination fill factor;	
			MProp.LamType=0;			// type of lamination;
			MProp.BHpoints=0;
			q[0]=NULL;
		}

		if( _strnicmp(q,"<blockname>",10)==0){
			v=StripKey(s);
			k=strlen(v);
			for(i=0;i<k;i++)
				if(v[i]=='\"'){
					v=v+i+1;
					i=k;
				}
			k=strlen(v);
			if(k>0) for(i=k-1;i>=0;i--){
				if(v[i]=='\"'){
					v[i]=0;
					i=-1;
				}
			}
			MProp.BlockName=v;
			q[0]=NULL;
		}

		if( _strnicmp(q,"<mu_x>",6)==0){
		   v=StripKey(s);
		   sscanf(v,"%lf",&MProp.mu_x);
		   q[0]=NULL;
		}
		
		if( _strnicmp(q,"<mu_y>",6)==0){
		   v=StripKey(s);
		   sscanf(v,"%lf",&MProp.mu_y);
		   q[0]=NULL;
		}	

		if( _strnicmp(q,"<H_c>",5)==0){
		   v=StripKey(s);
		   sscanf(v,"%lf",&MProp.H_c);
		   q[0]=NULL;
		}	

		if( _strnicmp(q,"<J_re>",6)==0){
		   v=StripKey(s);
		   sscanf(v,"%lf",&MProp.Jr);
		   q[0]=NULL;
		}	
		
		if( _strnicmp(q,"<J_im>",6)==0){
		   v=StripKey(s);
		   sscanf(v,"%lf",&MProp.Ji);
		   q[0]=NULL;
		}	

		if( _strnicmp(q,"<sigma>",7)==0){
		   v=StripKey(s);
		   sscanf(v,"%lf",&MProp.Cduct);
		   q[0]=NULL;
		}	
		
		if( _strnicmp(q,"<phi_h>",7)==0){
		   v=StripKey(s);
		   sscanf(v,"%lf",&MProp.Theta_h);
		   q[0]=NULL;
		}	
		
		if( _strnicmp(q,"<d_lam>",7)==0){
		   v=StripKey(s);
		   sscanf(v,"%lf",&MProp.Lam_d);
		   q[0]=NULL;
		}	
	
		if( _strnicmp(q,"<LamFill>",8)==0){
		   v=StripKey(s);
		   sscanf(v,"%lf",&MProp.LamFill);
		   q[0]=NULL;
		}

		if( _strnicmp(q,"<LamType>",9)==0){
		   v=StripKey(s);
		   sscanf(v,"%i",&MProp.LamType);
		   q[0]=NULL;
		}	

		if( _strnicmp(q,"<BHPoints>",10)==0){
			v=StripKey(s);
			sscanf(v,"%i",&MProp.BHpoints);
			if (MProp.BHpoints>0)
			{
				MProp.BHdata=(CComplex *)calloc(MProp.BHpoints,sizeof(CComplex));
				for(j=0;j<MProp.BHpoints;j++){
					fgets(s,1024,fp);
					sscanf(s,"%lf	%lf",&MProp.BHdata[j].re,&MProp.BHdata[j].im);
				}
			}
			q[0]=NULL;

		}

		if( _strnicmp(q,"<endblock>",9)==0){
			if (MProp.BHpoints>0) MProp.GetSlopes();
			blockproplist.Add(MProp);
			MProp.BHpoints=0;
			MProp.BHdata=NULL;
			MProp.slope=NULL;
			q[0]=NULL;
		}

		// Circuit Properties
		if( _strnicmp(q,"<begincircuit>",14)==0){	
			CProp.CircName="New Circuit";
			CProp.dVolts=0.;
			CProp.Amps=0.;
			CProp.CircType=0;
			q[0]=NULL;
		}

		if( _strnicmp(q,"<circuitname>",13)==0){
			v=StripKey(s);
			k=strlen(v);
			for(i=0;i<k;i++)
				if(v[i]=='\"'){
					v=v+i+1;
					i=k;
				}
			k=strlen(v);
			if(k>0) for(i=k-1;i>=0;i--){
				if(v[i]=='\"'){
					v[i]=0;
					i=-1;
				}
			}
			CProp.CircName=v;
			q[0]=NULL;
		}
		
		if( _strnicmp(q,"<voltgradient_re>",17)==0){
			double inval;
		   v=StripKey(s);
		   sscanf(v,"%lf",&inval);
		   CProp.dVolts+=inval;
		   q[0]=NULL;
		}

		if( _strnicmp(q,"<totalamps_re>",14)==0){
			double inval;
		   v=StripKey(s);
		   sscanf(v,"%lf",&inval);
		   CProp.Amps+=inval;
		   q[0]=NULL;
		}

		if( _strnicmp(q,"<voltgradient_im>",17)==0){
			double inval;
		   v=StripKey(s);
		   sscanf(v,"%lf",&inval);
		   CProp.dVolts+=(I*inval);
		   q[0]=NULL;
		}

		if( _strnicmp(q,"<totalamps_re>",14)==0){
			double inval;
		   v=StripKey(s);
		   sscanf(v,"%lf",&inval);
		   CProp.Amps+=(I*inval);
		   q[0]=NULL;
		}
		
		if( _strnicmp(q,"<circuittype>",13)==0){
		   v=StripKey(s);
		   sscanf(v,"%i",&CProp.CircType);
		   q[0]=NULL;
		}
		
		if( _strnicmp(q,"<endcircuit>",12)==0){
			circproplist.Add(CProp);
			q[0]=NULL;
		}

		// Points list;
		if(_strnicmp(q,"[numpoints]",11)==0){
			v=StripKey(s);
			sscanf(v,"%i",&k);
			for(i=0;i<k;i++)
			{
				fgets(s,1024,fp);
				sscanf(s,"%lf	%lf	%i\n",&node.x,&node.y,&t);
				node.BoundaryMarker=t-1;
				nodelist.Add(node);
			}
			q[0]=NULL;
		}

		// read in segment list
		if(_strnicmp(q,"[numsegments]",13)==0){
			v=StripKey(s);
			sscanf(v,"%i",&k);
			for(i=0;i<k;i++)
			{
				fgets(s,1024,fp);
				sscanf(s,"%i	%i	%lf %i	%i\n",&segm.n0,&segm.n1,&segm.MaxSideLength,&t,&segm.Hidden);
				segm.BoundaryMarker=t-1;
				linelist.Add(segm);
			}
			q[0]=NULL;
		}

		// read in arc segment list
		if(_strnicmp(q,"[numarcsegments]",13)==0){
			v=StripKey(s);
			sscanf(v,"%i",&k);
			for(i=0;i<k;i++)
			{
				fgets(s,1024,fp);
				sscanf(s,"%i	%i	%lf	%lf %i	%i\n",&asegm.n0,&asegm.n1,
					&asegm.ArcLength,&asegm.MaxSideLength,&t,&asegm.Hidden);
				asegm.BoundaryMarker=t-1;
				arclist.Add(asegm);
			}
			q[0]=NULL;
		}


		// read in list of holes;
		if(_strnicmp(q,"[numholes]",13)==0){
			v=StripKey(s);
			sscanf(v,"%i",&k);
			if(k>0)
			{
				blk.BlockType=-1;
				blk.MaxArea=0;
				for(i=0;i<k;i++)
				{	
					fgets(s,1024,fp);
					sscanf(s,"%lf	%lf\n",&blk.x,&blk.y);
				//	blocklist.Add(blk);
				//  don't add holes to the list
				//  of block labels because it messes up the
				//  number of block labels.
				}
			}
			q[0]=NULL;
		}

		// read in regional attributes
		if(_strnicmp(q,"[numblocklabels]",13)==0){
			v=StripKey(s);
			sscanf(v,"%i",&k);
			for(i=0;i<k;i++)
			{
				fgets(s,1024,fp);
				sscanf(s,"%lf	%lf	%i	%lf	%i	%lf\n",&blk.x,&blk.y,
					&blk.BlockType,&blk.MaxArea,&blk.InCircuit,&blk.MagDir);
				if (blk.MaxArea<0) blk.MaxArea=0;
				else blk.MaxArea=PI*blk.MaxArea*blk.MaxArea/4.;
				blk.BlockType-=1;
				blk.InCircuit-=1;
				blocklist.Add(blk);
			}		
			q[0]=NULL;
		}

		if(_strnicmp(q,"[solution]",10)==0){
			flag=TRUE;
			q[0]=NULL;
		}
	}	

	// read in meshnodes;
	fscanf(fp,"%i\n",&k);
	meshnode.SetSize(k);
	for(i=0;i<k;i++)
	{
		fgets(s,1024,fp);
		if (Frequency>0)
			sscanf(s,"%lf	%lf	%lf	%lf",&mnode.x,&mnode.y,
				&mnode.A.re,&mnode.A.im);
		else{
			sscanf(s,"%lf	%lf	%lf",&mnode.x,&mnode.y,&mnode.A.re);
			mnode.A.im=0;
		}
		meshnode.SetAt(i,mnode);
	}

	// read in elements;
	fscanf(fp,"%i\n",&k);
	meshelem.SetSize(k);
	for(i=0;i<k;i++)
	{
		fgets(s,1024,fp);
		if (Frequency==0)
			sscanf(s,"%i	%i	%i	%i	%lf	%lf",&elm.p[0],&elm.p[1],&elm.p[2],
				&elm.lbl,&elm.mu1,&elm.mu2);
		else{
			sscanf(s,"%i	%i	%i	%i",&elm.p[0],&elm.p[1],&elm.p[2],
				&elm.lbl);
			
			elm.mu1=0;
			elm.mu2=0;
		}
		elm.blk=blocklist[elm.lbl].BlockType;
		meshelem.SetAt(i,elm);
	}	

	// read in circuit data;
	fscanf(fp,"%i\n",&k);
	for(i=0;i<k;i++){
		fgets(s,1024,fp);
		if (Frequency==0){
			sscanf(s,"%i	%lf",&j,&zr);
			circproplist[i].Case=j;
			if (j==0) circproplist[i].dVolts=zr;
			else circproplist[i].J=zr;
		}
		else{
			sscanf(s,"%i	%lf	%lf",&j,&zr,&zi);
			circproplist[i].Case=j;
			if (j==0) circproplist[i].dVolts=zr + I*zi;
			else circproplist[i].J=zr + I*zi;
		}
	}

	fclose(fp);

	// Find flux density in each element;
	for(i=0;i<meshelem.GetSize();i++) GetElementB(meshelem[i]);
	
	// Find extreme values of A;
	A_Low=meshnode[0].A.re; A_High=meshnode[0].A.re;
	for(i=1;i<meshnode.GetSize();i++)
	{
		if (meshnode[i].A.re>A_High) A_High=meshnode[i].A.re;
		if (meshnode[i].A.re<A_Low)  A_Low =meshnode[i].A.re;
		
		if(Frequency!=0)
		{
			if (meshnode[i].A.im<A_Low)  A_Low =meshnode[i].A.im;
			if (meshnode[i].A.im>A_High) A_High=meshnode[i].A.im;
		}
	}
	// save default values for extremes of A
	A_lb=A_Low;
	A_ub=A_High;

	if(Frequency>0){ // compute frequency-dependent permeabilities;
		
		CComplex deg45; deg45=1+I;
		CComplex K,lag,halflag;
		double ds;
		double w=2.*PI*Frequency;

		for(k=0;k<blockproplist.GetSize();k++){
		lag=exp(-I*blockproplist[k].Theta_h*PI/180.);
		if (blockproplist[k].LamType==0){
			blockproplist[k].mu_fdx.Set(blockproplist[k].mu_x,0);
			blockproplist[k].mu_fdy.Set(blockproplist[k].mu_y,0);
			blockproplist[k].mu_fdx=Mult(blockproplist[k].mu_fdx,lag);
			blockproplist[k].mu_fdy=Mult(blockproplist[k].mu_fdy,lag);
			
			if(blockproplist[k].Lam_d!=0){
				halflag=exp(-I*blockproplist[k].Theta_h*PI/360.);
				ds=sqrt(2./(0.4*PI*w*blockproplist[k].Cduct*blockproplist[k].mu_x));
				K=halflag*deg45*blockproplist[k].Lam_d*0.001/(2.*ds);
				blockproplist[k].mu_fdx=(blockproplist[k].mu_fdx*tanh(K)/K)*
					blockproplist[k].LamFill;
	
				ds=sqrt(2./(0.4*PI*w*blockproplist[k].Cduct*blockproplist[k].mu_y));
				K=halflag*deg45*blockproplist[k].Lam_d*0.001/(2.*ds);
				blockproplist[k].mu_fdy=(blockproplist[k].mu_fdy*tanh(K)/K)*
					blockproplist[k].LamFill;
			}
		}
		if (blockproplist[k].LamType==1){
				x=blockproplist[k].LamFill;
				K=lag*blockproplist[k].mu_x;
				blockproplist[k].mu_fdx = K*x + (1.-x);
				blockproplist[k].mu_fdy = K/(x + K*(1.-x));
		}
		if (blockproplist[k].LamType==2){
				x=blockproplist[k].LamFill;
				K=lag*blockproplist[k].mu_y;
				blockproplist[k].mu_fdy = K*x + (1.-x);
				blockproplist[k].mu_fdx = K/(x + K*(1.-x));
		}
		
	}}
	
	// element centroids and radii;
	for(i=0;i<meshelem.GetSize();i++)
	{
		meshelem[i].ctr=Ctr(i);
		for(j=0,meshelem[i].rsqr=0;j<3;j++)
		{
			b=sqr(meshnode[meshelem[i].p[j]].x-meshelem[i].ctr.re)+
			  sqr(meshnode[meshelem[i].p[j]].y-meshelem[i].ctr.im);
			if(b>meshelem[i].rsqr) meshelem[i].rsqr=b;
		}
	}

	// build list of elements connected to each node;
	// allocate connections list;
	NumList=(int *)calloc(meshnode.GetSize(),sizeof(int));
	ConList=(int **)calloc(meshnode.GetSize(),sizeof(int *));
	// find out number of connections to each node;
	for(i=0;i<meshelem.GetSize();i++)
		for(j=0;j<3;j++)
			NumList[meshelem[i].p[j]]++;
	// allocate space for connections lists;
	for(i=0;i<meshnode.GetSize();i++)
		ConList[i]=(int *)calloc(NumList[i],sizeof(int));
	// build list;
	for(i=0;i<meshnode.GetSize();i++) NumList[i]=0;
	for(i=0;i<meshelem.GetSize();i++)
		for(j=0;j<3;j++){
			k=meshelem[i].p[j];
			ConList[k][NumList[k]]=i;
			NumList[k]++;
		}

	// Find extreme values of B;
	Br_Low=sqrt(sqr(meshelem[0].B1.re) + sqr(meshelem[0].B2.re));
	Br_High=Br_Low;
	Bi_Low=sqrt(sqr(meshelem[0].B1.im)+ sqr(meshelem[0].B2.im));
	Bi_High=Bi_Low;
	B_Low=sqrt(Br_Low*Br_Low + Bi_Low*Bi_Low);
	B_High=B_Low;
	for(i=0;i<meshelem.GetSize();i++)
	{
		GetNodalB(meshelem[i].b1,meshelem[i].b2,meshelem[i]);
		for(j=0;j<3;j++){
			br=sqrt(sqr(meshelem[i].b1[j].re) + 
				    sqr(meshelem[i].b2[j].re));
			bi=sqrt(sqr(meshelem[i].b1[j].im) +
				    sqr(meshelem[i].b2[j].im));
			b=sqrt(br*br+bi*bi);
	
			if(b>B_High) B_High=b; if(b<B_Low) B_Low=b;
			if(br>Br_High) Br_High=br; if(br<Br_Low) Br_Low=br;
			if(bi>Bi_High) Bi_High=bi; if(bi<Bi_Low) Bi_Low=bi;
		}
	}

	B_lb=B_Low; B_ub=B_High;

	// compute total resulting current for circuits with a defined
	// voltage gradient;  Need this to display circuit results & impedance.
	for(i=0;i<circproplist.GetSize();i++)
	{
		CComplex Jelm[3],Aelm[3];
		double a;
		
		if(circproplist[i].CircType!=0)
		for(j=0,circproplist[i].Amps=0.;j<meshelem.GetSize();j++)
		{
			if(blocklist[meshelem[j].lbl].InCircuit==i)
			{
				GetJA(j,Jelm,Aelm);
				a=ElmArea(j)*sqr(LengthConv[LengthUnits]);
				for(k=0;k<3;k++) circproplist[i].Amps+=a*Jelm[k]/3;
			}
		}
	}


	// Build adjacency information for each element.
	// This info is required to find points within the
	// mesh quickly, instead of by a brute-force search.
	BuildElemMap(&meshnode, &meshelem, NumList, ConList);
	
	FirstDraw=TRUE;
	 

	return TRUE;
}

/*  Original slooow version 
int CFemmviewDoc::InTriangle(double x, double y)
{
	int i;
	double z;
	BOOL InFlag;

	for(i=0;i<meshelem.GetSize();i++)
	{
		z=(meshelem[i].ctr.re-x)*(meshelem[i].ctr.re-x) +
			(meshelem[i].ctr.im-y)*(meshelem[i].ctr.im-y);
		if(z>meshelem[i].rsqr) InFlag=FALSE;
		else InFlag=InTriangleTest(x,y,i); 

		if(InFlag==TRUE) return i;
	}

	return (-1);
}
*/

// Si hang improved version of InTriangle()
int CFemmviewDoc::InTriangle(double x, double y)
{
	CMeshNode searchpoint;
	TriEdge searchtri;
	
	searchpoint.x = x;
	searchpoint.y = y;
	searchtri.ele = 0; // Choose a Elem Arbitrarily.
	
	enum LocateResult lres = Locate(searchpoint, &searchtri);
	if(lres != eOutSide) return searchtri.ele;
	else return (-1);
}

BOOL CFemmviewDoc::GetPointValues(double x, double y, CPointVals &u)
{
	int k;
	k=InTriangle(x,y);
	if (k<0) return FALSE;
	GetPointValues(x,y,k,u);
	return TRUE;
}

BOOL CFemmviewDoc::GetPointValues(double x, double y, int k, CPointVals &u)
{
	int i,j,n[3];
	double a[3],b[3],c[3],da,ravg;
	
	for(i=0;i<3;i++) n[i]=meshelem[k].p[i];
	a[0]=meshnode[n[1]].x * meshnode[n[2]].y - meshnode[n[2]].x * meshnode[n[1]].y;
	a[1]=meshnode[n[2]].x * meshnode[n[0]].y - meshnode[n[0]].x * meshnode[n[2]].y;
	a[2]=meshnode[n[0]].x * meshnode[n[1]].y - meshnode[n[1]].x * meshnode[n[0]].y;
	b[0]=meshnode[n[1]].y - meshnode[n[2]].y;
	b[1]=meshnode[n[2]].y - meshnode[n[0]].y;
	b[2]=meshnode[n[0]].y - meshnode[n[1]].y;	
	c[0]=meshnode[n[2]].x - meshnode[n[1]].x;
	c[1]=meshnode[n[0]].x - meshnode[n[2]].x;
	c[2]=meshnode[n[1]].x - meshnode[n[0]].x;
	da=(b[0]*c[1]-b[1]*c[0]);
	ravg=LengthConv[LengthUnits]*
		(meshnode[n[0]].x + meshnode[n[1]].x + meshnode[n[2]].x)/3.;
	
	GetPointB(x,y,u.B1,u.B2,meshelem[k]);
	
	if (Frequency==0){
		u.A.Set(0,0);
		if(ProblemType==0){
			for(i=0;i<3;i++)
				u.A.re+=meshnode[n[i]].A.re*(a[i]+b[i]*x+c[i]*y)/(da);
		}
		else{
		/* Old way that I interpolated potential in axi case:
			// interpolation from A from nodal points.
			// note that the potential that's actually stored
			// for axisymmetric problems is 2*Pi*r*A, so divide
			// by nodal r to get 2*Pi*A at the nodes.  Linearly
			// interpolate this, then multiply by r at the point
			// of interest to get back to 2*Pi*r*A.
			for(i=0,rp=0;i<3;i++){
				r=meshnode[n[i]].x;
				rp+=meshnode[n[i]].x*(a[i]+b[i]*x+c[i]*y)/da;
				if (r>1.e-6) u.A.re+=meshnode[n[i]].A.re*
					(a[i]+b[i]*x+c[i]*y)/(r*da);
			}
			u.A.re*=rp;
		*/

			
			// a ``smarter'' interpolation.  One based on A can't
			// represent constant flux density very well.
			// This works, but I should re-write it in a more
			// efficient form--it's doing a lot of the work
			// twice, because the a-b-c stuff that's already
			// been computed is ignored.
			double v[6];
			double R[3];
			double Z[3];
			double p,q;

			for(i=0;i<3;i++){
				R[i]=meshnode[n[i]].x;
				Z[i]=meshnode[n[i]].y;
			}

			// corner nodes
			v[0]=meshnode[n[0]].A.re;
			v[2]=meshnode[n[1]].A.re;
			v[4]=meshnode[n[2]].A.re;

			// construct values for mid-side nodes;
			if ((R[0]<1.e-06) && (R[1]<1.e-06))
			v[1]=(v[0]+v[2])/2.;
			else
			v[1]=(R[1]*(3.*v[0] + v[2]) + R[0]*(v[0] + 3.*v[2]))/
				 (4.*(R[0] + R[1]));

			if ((R[1]<1.e-06) && (R[2]<1.e-06))
			v[3]=(v[2]+v[4])/2.;
			else
			v[3]=(R[2]*(3.*v[2] + v[4]) + R[1]*(v[2] + 3.*v[4]))/
				 (4.*(R[1] + R[2]));

			if ((R[2]<1.e-06) && (R[0]<1.e-06))
			v[5]=(v[4]+v[0])/2.;
			else
			v[5]=(R[0]*(3.*v[4] + v[0]) + R[2]*(v[4] + 3.*v[0]))/
				 (4.*(R[2] + R[0]));

			// compute location in element transformed onto
			// a unit triangle;
			p=(b[1]*x+c[1]*y + a[1])/da;
			q=(b[2]*x+c[2]*y + a[2])/da;

			// now, interpolate to get potential...
			u.A.re = v[0] - p*(3.*v[0] - 4.*v[1] + v[2]) +
				     2.*p*p*(v[0] - 2.*v[1] + v[2]) - 
				     q*(3.*v[0] + v[4] - 4.*v[5]) + 
					 2.*q*q*(v[0] + v[4] - 2.*v[5]) + 
			         4.*p*q*(v[0] - v[1] + v[3] - v[5]);

	/*		// "simple" way to do it...
			// problem is that this mucks up things
			// near the centerline, where things ought
			// to look pretty quadratic.
			for(i=0;i<3;i++)
				u.A.re+=meshnode[n[i]].A.re*(a[i]+b[i]*x+c[i]*y)/(da);
	*/
		}
		u.mu1.im=0; u.mu2.im=0;
		blockproplist[meshelem[k].blk].GetMu(u.B1.re,u.B2.re,u.mu1.re,u.mu2.re);
		u.H1=Div(u.B1,u.mu1.re*PI*4.e-07);
		u.H2=Div(u.B2,u.mu2.re*PI*4.e-07);
		u.Je.Set(0,0);
		u.Js.Set(blockproplist[meshelem[k].blk].Jr,0);
		j=blocklist[meshelem[k].lbl].InCircuit;
		if(j>=0){
			if(circproplist[j].Case==0){
				if (ProblemType==0)
					u.Js-=blockproplist[meshelem[k].blk].Cduct*
						  circproplist[j].dVolts;
				else
					u.Js-=blockproplist[meshelem[k].blk].Cduct*
					      circproplist[j].dVolts/ravg;
			}
			else u.Js+=circproplist[j].J;
		}			
		u.c=blockproplist[meshelem[k].blk].Cduct;
		u.E=blockproplist[meshelem[k].blk].DoEnergy(u.B1.re,u.B2.re);
		u.Ph=0;
		u.Pe=0;
		return TRUE;
	}
		
	if(Frequency>0){
		u.A.Set(0,0);
		if(ProblemType==0)
		{
			for(i=0;i<3;i++)
				u.A+=meshnode[n[i]].A*(a[i]+b[i]*x+c[i]*y)/(da);
		}
		else{
			CComplex v[6];
			double R[3];
			double Z[3];
			double p,q;

			for(i=0;i<3;i++){
				R[i]=meshnode[n[i]].x;
				Z[i]=meshnode[n[i]].y;
			}

			// corner nodes
			v[0]=meshnode[n[0]].A;
			v[2]=meshnode[n[1]].A;
			v[4]=meshnode[n[2]].A;

			// construct values for mid-side nodes;
			if ((R[0]<1.e-06) && (R[1]<1.e-06))
			v[1]=(v[0]+v[2])/2.;
			else
			v[1]=(R[1]*(3.*v[0] + v[2]) + R[0]*(v[0] + 3.*v[2]))/
				 (4.*(R[0] + R[1]));

			if ((R[1]<1.e-06) && (R[2]<1.e-06))
			v[3]=(v[2]+v[4])/2.;
			else
			v[3]=(R[2]*(3.*v[2] + v[4]) + R[1]*(v[2] + 3.*v[4]))/
				 (4.*(R[1] + R[2]));

			if ((R[2]<1.e-06) && (R[0]<1.e-06))
			v[5]=(v[4]+v[0])/2.;
			else
			v[5]=(R[0]*(3.*v[4] + v[0]) + R[2]*(v[4] + 3.*v[0]))/
				 (4.*(R[2] + R[0]));

			// compute location in element transformed onto
			// a unit triangle;
			p=(b[1]*x+c[1]*y + a[1])/da;
			q=(b[2]*x+c[2]*y + a[2])/da;

			// now, interpolate to get potential...
			u.A = v[0] - p*(3.*v[0] - 4.*v[1] + v[2]) +
				     2.*p*p*(v[0] - 2.*v[1] + v[2]) - 
				     q*(3.*v[0] + v[4] - 4.*v[5]) + 
					 2.*q*q*(v[0] + v[4] - 2.*v[5]) + 
			         4.*p*q*(v[0] - v[1] + v[3] - v[5]);
		}

		u.mu1=blockproplist[meshelem[k].blk].mu_fdx;
		u.mu2=blockproplist[meshelem[k].blk].mu_fdy;
		u.H1=u.B1/(u.mu1*PI*4.e-07);
		u.H2=u.B2/(u.mu2*PI*4.e-07);
		u.Js.Set(blockproplist[meshelem[k].blk].Jr,
				 blockproplist[meshelem[k].blk].Ji);
		j=blocklist[meshelem[k].lbl].InCircuit;
		if(j>=0){
			if(circproplist[j].Case==0){
				if (ProblemType==0)
					u.Js-=blockproplist[meshelem[k].blk].Cduct*
						  circproplist[j].dVolts;
				else
					u.Js-=blockproplist[meshelem[k].blk].Cduct*
					      circproplist[j].dVolts/ravg;
			}
			else u.Js+=circproplist[j].J;
		}	
		u.c=blockproplist[meshelem[k].blk].Cduct;
		if (blockproplist[meshelem[k].blk].Lam_d!=0) u.c=0;
		
		u.Je=-I*Frequency*2.*PI*u.c*u.A;

		if(ProblemType!=0){
			if(x!=0)
				u.Je/=(2.*PI*x*LengthConv[LengthUnits]);
			else u.Je=0;
		}
		u.E=0.5*(u.B1.re*u.H1.re+u.B2.re*u.H2.re);
		CComplex z;
		z=(u.H1*u.B1.Conj()) + (u.H2*u.B2.Conj());
		u.E=0.25*z.re;
		u.Ph=Frequency*PI*z.im;
		z=u.Js + u.Je;
		u.Pe=1.e06*(z.Abs()*z.Abs())/(u.c*2.);
		return TRUE;
	}	

	return FALSE;
}

void CFemmviewDoc::GetPointB(double x, double y, CComplex &B1, CComplex &B2,
							 CElement &elm)
{
	// elm is a reference to the element that contains the point of interest.
	int i,n[3];
	double da,a[3],b[3],c[3];

	if(Smooth==FALSE){
		B1=elm.B1;
		B2=elm.B2;
		return;
	}

	for(i=0;i<3;i++) n[i]=elm.p[i];
	a[0]=meshnode[n[1]].x * meshnode[n[2]].y - meshnode[n[2]].x * meshnode[n[1]].y;
	a[1]=meshnode[n[2]].x * meshnode[n[0]].y - meshnode[n[0]].x * meshnode[n[2]].y;
	a[2]=meshnode[n[0]].x * meshnode[n[1]].y - meshnode[n[1]].x * meshnode[n[0]].y;
	b[0]=meshnode[n[1]].y - meshnode[n[2]].y;
	b[1]=meshnode[n[2]].y - meshnode[n[0]].y;
	b[2]=meshnode[n[0]].y - meshnode[n[1]].y;	
	c[0]=meshnode[n[2]].x - meshnode[n[1]].x;
	c[1]=meshnode[n[0]].x - meshnode[n[2]].x;
	c[2]=meshnode[n[1]].x - meshnode[n[0]].x;
	da=(b[0]*c[1]-b[1]*c[0]);

	B1.Set(0,0);
	B2.Set(0,0);
	for(i=0;i<3;i++){
		B1+=(elm.b1[i]*(a[i]+b[i]*x+c[i]*y)/da);
		B2+=(elm.b2[i]*(a[i]+b[i]*x+c[i]*y)/da);
	}
}

void CFemmviewDoc::GetNodalB(CComplex *b1, CComplex *b2,CElement &elm)
{
	// elm is a reference to the element that contains the point of interest.
	CComplex p;
	CComplex tn,bn,bt,btu,btv,u1,u2,v1,v2;
	int i,j,k,l,q,m,pt,nxt;
	double r,R,z;
	CElement *e;
	int flag;

	// find nodal values of flux density via a patch method.
	for(i=0;i<3;i++){

		k=elm.p[i];
		p.Set(meshnode[k].x,meshnode[k].y);
		b1[i].Set(0,0);
		b2[i].Set(0,0);
		for(j=0,m=0;j<NumList[k];j++)
			if(elm.blk==meshelem[ConList[k][j]].blk) m++;
			else{
				if(Frequency==0){
					if ((elm.mu1==meshelem[ConList[k][j]].mu1) &&
						(elm.mu2==meshelem[ConList[k][j]].mu2) &&
						(blockproplist[elm.blk].H_c==
						 blockproplist[meshelem[ConList[k][j]].blk].H_c) &&
						(blocklist[elm.lbl].MagDir==
						 blocklist[meshelem[ConList[k][j]].lbl].MagDir)) m++;
				}
				else if ((blockproplist[elm.blk].mu_fdx==
						  blockproplist[meshelem[ConList[k][j]].blk].mu_fdx) &&
						 (blockproplist[elm.blk].mu_fdy==
						  blockproplist[meshelem[ConList[k][j]].blk].mu_fdy)) m++;
			}
		
		if(m==NumList[k])
		{
			for(j=0,R=0;j<NumList[k];j++)
			{
				m=ConList[k][j];
				z=1./abs(p-Ctr(m));
				R+=z;
				b1[i]+=(z*meshelem[m].B1);
				b2[i]+=(z*meshelem[m].B2);
			}
			b1[i]/=R;
			b2[i]/=R;
		}
		else{
			R=0; v1=0; v2=0;
			
			//scan ccw for an interface...
			e=&elm;
			for(q=0;q<NumList[k];q++)
			{
				//find ccw side of the element;
				for(j=0;j<3;j++) if(e->p[j]==k) pt=j;
				pt--;
				if(pt<0) pt=2;
				pt=e->p[pt];

				//scan to find element adjacent to this side;
				for(j=0,nxt=-1;j<NumList[k];j++){
					if(&meshelem[ConList[k][j]]!=e){
						for(l=0;l<3;l++)
							if(meshelem[ConList[k][j]].p[l]==pt)
							nxt=ConList[k][j];
					}
				}
				if(nxt==-1) q=NumList[k];
				else if(elm.blk!=meshelem[nxt].blk){
					// we have found two elements on either side of the interface
					// now, we take contribution from B at the center of the 
					// interface side
					tn.Set(meshnode[pt].x-meshnode[k].x,
						  meshnode[pt].y-meshnode[k].y);
					r=(meshnode[pt].x+meshnode[k].x)*LengthConv[LengthUnits]/2.;
					bn=(meshnode[pt].A-meshnode[k].A)/
					   (abs(tn)*LengthConv[LengthUnits]);
					if(ProblemType==1){
						bn/=(-2.*PI*r);
					}
					z=0.5/abs(tn);
					tn/=abs(tn);

					// for the moment, kludge with bt...
					bt=e->B1*tn.re + e->B2*tn.im;
										
					R+=z;
					b1[i]+=(z*tn.re*bt);
					b2[i]+=(z*tn.im*bt);
					b1[i]+=(z*tn.im*bn);
					b2[i]+=(-z*tn.re*bn);
					v1=tn;
					q=NumList[k];
				}
				else e=&meshelem[nxt];
			}

			//scan cw for an interface...
			e=&elm;
			for(q=0;q<NumList[k];q++)
			{
				//find cw side of the element;
				for(j=0;j<3;j++) if(e->p[j]==k) pt=j;
				pt++;
				if(pt>2) pt=0;
				pt=e->p[pt];

				//scan to find element adjacent to this side;
				for(j=0,nxt=-1;j<NumList[k];j++){
					if(&meshelem[ConList[k][j]]!=e){
						for(l=0;l<3;l++)
							if(meshelem[ConList[k][j]].p[l]==pt)
							nxt=ConList[k][j];
					}
				}
				if (nxt==-1) q=NumList[k];
				else if(elm.blk!=meshelem[nxt].blk){
					// we have found two elements on either side of the interface
					// now, we take contribution from B at the center of the 
					// interface side
					tn.Set(meshnode[pt].x-meshnode[k].x,
						  meshnode[pt].y-meshnode[k].y);
					r=(meshnode[pt].x+meshnode[k].x)*LengthConv[LengthUnits]/2.;
					bn=(meshnode[pt].A-meshnode[k].A)/
						(abs(tn)*LengthConv[LengthUnits]);
					if(ProblemType==1){
						bn/=(-2.*PI*r);
					}
					z=0.5/abs(tn);
					tn/=abs(tn);
				
					// for the moment, kludge with bt...
					bt=e->B1*tn.re + e->B2*tn.im;
						
					R+=z;
					b1[i]+=(z*tn.re*bt);
					b2[i]+=(z*tn.im*bt);
					b1[i]+=(z*tn.im*bn);
					b2[i]+=(-z*tn.re*bn);
					v2=tn;	
					q=NumList[k];
				}
				else e=&meshelem[nxt];
			}
			b1[i]/=R;
			b2[i]/=R;

			// check to see if angle of corner is too sharp to apply
			// this rule; really only does right if the interface is flat;
			flag=FALSE;	
			// if there is only one edge, approx is ok;
			if ((abs(v1)<0.9) || (abs(v2)<0.9)) flag=TRUE;
			// if the interfaces make less than a 10 degree angle, things are ok;
			if ( (-v1.re*v2.re-v1.im*v2.im) > 0.985) flag=TRUE;
			
			// Otherwise, punt...
			if(flag==FALSE){
				bn=0;
				k=elm.p[i];
				for(j=0;j<NumList[k];j++)
				{
					if(elm.blk==meshelem[ConList[k][j]].blk){
						m=ConList[k][j];
						bt.re=sqrt(meshelem[m].B1.re*meshelem[m].B1.re +
							meshelem[m].B2.re*meshelem[m].B2.re);
						bt.im=sqrt(meshelem[m].B1.im*meshelem[m].B1.im +
							meshelem[m].B2.im*meshelem[m].B2.im);
						if(bt.re>bn.re) bn.re=bt.re;
						if(bt.im>bn.im) bn.im=bt.im;
					}
				}
			
				R=sqrt(elm.B1.re*elm.B1.re + elm.B2.re*elm.B2.re);
				if(R!=0)
				{
					b1[i].re=bn.re/R * elm.B1.re;
					b2[i].re=bn.re/R * elm.B2.re;
				}
				else{
					b1[i].re=0;
					b2[i].re=0;
				}

				R=sqrt(elm.B1.im*elm.B1.im + elm.B2.im*elm.B2.im);
				if(R!=0)
				{
					b1[i].im=bn.im/R * elm.B1.im;
					b2[i].im=bn.im/R * elm.B2.im;
				}
				else{
					b1[i].im=0;
					b2[i].im=0;
				}
			}
		}

		// check to see if the point has a point current; if so, just
		// use element average values;
		if (nodeproplist.GetSize()!=0)
		for(j=0;j<nodelist.GetSize();j++)
		{
			if (abs(p-(nodelist[j].x+nodelist[j].y*I))<1.e-08)
			if(nodelist[j].BoundaryMarker>=0)
			{	
				if ((nodeproplist[nodelist[j].BoundaryMarker].Jr!=0) ||
					(nodeproplist[nodelist[j].BoundaryMarker].Ji!=0))
				{
					b1[i]=elm.B1;
					b2[i]=elm.B2;	
				}
			}
		}
		

		//check for special case of node on r=0 axisymmetric; set Br=0;
		if ((fabs(p.re)<1.e-06) && (ProblemType==1)) b1[i].Set(0.,0);
	}
}

void CFemmviewDoc::GetElementB(CElement &elm)
{
	int i,n[3];
	double b[3],c[3],da;

	for(i=0;i<3;i++) n[i]=elm.p[i];

	b[0]=meshnode[n[1]].y - meshnode[n[2]].y;
	b[1]=meshnode[n[2]].y - meshnode[n[0]].y;
	b[2]=meshnode[n[0]].y - meshnode[n[1]].y;	
	c[0]=meshnode[n[2]].x - meshnode[n[1]].x;
	c[1]=meshnode[n[0]].x - meshnode[n[2]].x;
	c[2]=meshnode[n[1]].x - meshnode[n[0]].x;
	da=(b[0]*c[1]-b[1]*c[0]);

	if (ProblemType==0){
		elm.B1=0;
		elm.B2=0;
		for(i=0;i<3;i++)
		{
			elm.B1+=meshnode[n[i]].A*c[i]/(da*LengthConv[LengthUnits]);
			elm.B2-=meshnode[n[i]].A*b[i]/(da*LengthConv[LengthUnits]);
		}
		return;
	}
	else{
	/*	
		// Old way that B at centerpoint of an element
		// was obtained.  Done on the basis of a direct A solution.
		// Problem is, solution has been changed to one that
		// doesn't interpolat this way anymore, so need to change
		// the way this is obtained.
		double r[3],R;
		CComplex A[3],Ao;
		for(i=0;i<3;i++){
			r[i]=meshnode[n[i]].x*LengthConv[LengthUnits];
			if (r[i]>1.e-6) A[i]=meshnode[n[i]].A/(2.*PI*r[i]);
			else A[i]=0;
		}
		R=(r[0]+r[1]+r[2])/3.;
		elm.B1=0;
		elm.B2=0;
		Ao=(A[0]+A[1]+A[2])/3.;

		for(i=0;i<3;i++)
		{
			elm.B1 -= A[i]*c[i]/(da*LengthConv[LengthUnits]);
			elm.B2 += A[i]*b[i]/(da*LengthConv[LengthUnits]);
		}
		elm.B2=(elm.B2 + Ao/R); 
*/
			CComplex v[6],dp,dq;
			double R[3],Z[3],r;

			for(i=0,r=0;i<3;i++){
				R[i]=meshnode[n[i]].x;
				Z[i]=meshnode[n[i]].y;
				r+=R[i]/3.;
			}

			// corner nodes
			v[0]=meshnode[n[0]].A;
			v[2]=meshnode[n[1]].A;
			v[4]=meshnode[n[2]].A;

			// construct values for mid-side nodes;
			if ((R[0]<1.e-06) && (R[1]<1.e-06))
			v[1]=(v[0]+v[2])/2.;
			else
			v[1]=(R[1]*(3.*v[0] + v[2]) + R[0]*(v[0] + 3.*v[2]))/
				 (4.*(R[0] + R[1]));

			if ((R[1]<1.e-06) && (R[2]<1.e-06))
			v[3]=(v[2]+v[4])/2.;
			else
			v[3]=(R[2]*(3.*v[2] + v[4]) + R[1]*(v[2] + 3.*v[4]))/
				 (4.*(R[1] + R[2]));

			if ((R[2]<1.e-06) && (R[0]<1.e-06))
			v[5]=(v[4]+v[0])/2.;
			else
			v[5]=(R[0]*(3.*v[4] + v[0]) + R[2]*(v[4] + 3.*v[0]))/
				 (4.*(R[2] + R[0]));

			// derivatives w.r.t. p and q:
			dp=(-v[0] + v[2] + 4.*v[3] - 4.*v[5])/3.;
			dq=(-v[0] - 4.*v[1] + 4.*v[3] + v[4])/3.;

			// now, compute flux.
			da*=2.*PI*r*LengthConv[LengthUnits]*LengthConv[LengthUnits];
			elm.B1=-(c[1]*dp+c[2]*dq)/da;
			elm.B2= (b[1]*dp+b[2]*dq)/da;

		return;
	}
}


void CFemmviewDoc::OnReload() 
{
	// TODO: Add your command handler code here
	CString pname = GetPathName();
	if(pname.GetLength()>0)
	{
		OnNewDocument();
		SetPathName(pname,FALSE);
		OnOpenDocument(pname);
	}
}

int CFemmviewDoc::ClosestNode(double x, double y)
{
	int i,j;
	double d0,d1;

	if(nodelist.GetSize()==0) return -1;

	j=0;
	d0=nodelist[0].GetDistance(x,y);
	for(i=0;i<nodelist.GetSize();i++){
		d1=nodelist[i].GetDistance(x,y);
		if(d1<d0){
			d0=d1;
			j=i;
		}
	}

	return j;
}

void CFemmviewDoc::GetLineValues(CXYPlot &p,int PlotType,int NumPlotPoints)
{
	double *q,z,u,dz;
	CComplex pt,n,t;
	int i,j,k,m,elm;
	CPointVals v;
	BOOL flag;

	q=(double *)calloc(contour.GetSize(),sizeof(double));
	for(i=1;i<contour.GetSize();i++)
	{
		z+=abs(contour[i]-contour[i-1]);
		q[i]=z;
	}
	dz=z/(NumPlotPoints-1);
	
	/*
		m_XYPlotType.AddString("Potential");
		m_XYPlotType.AddString("|B|        (Magnitude of flux density)");
		m_XYPlotType.AddString("B . n      (Normal flux density)");
		m_XYPlotType.AddString("B . t      (Tangential flux density)");
		m_XYPlotType.AddString("|H|        (Magnitude of field intensity)");
		m_XYPlotType.AddString("H . n      (Normal field intensity)");
		m_XYPlotType.AddString("H . t      (Tangential field intensity)");
		m_XYPlotType.AddString("J_eddy     
	*/

	if(Frequency==0){
		switch (PlotType)
		{
			case 0:
				p.Create(NumPlotPoints,2);
				if (ProblemType==0) strcpy(p.lbls[1],"Potential, Wb/m");
				else strcpy(p.lbls[1],"Flux, Wb");
				break;
			case 1:
				p.Create(NumPlotPoints,2);
				strcpy(p.lbls[1],"|B|, Tesla");
				break;
			case 2:
				p.Create(NumPlotPoints,2);
				strcpy(p.lbls[1],"B.n, Tesla");
				break;
			case 3:
				p.Create(NumPlotPoints,2);
				strcpy(p.lbls[1],"B.t, Tesla");
				break;
			case 4:
				p.Create(NumPlotPoints,2);
				strcpy(p.lbls[1],"|H|, Amp/m");
				break;
			case 5:
				p.Create(NumPlotPoints,2);
				strcpy(p.lbls[1],"H.n, Amp/m");
				break;
			case 6:
				p.Create(NumPlotPoints,2);
				strcpy(p.lbls[1],"H.t, Amp/m");
				break;
			default:
				p.Create(NumPlotPoints,2);
				break;
		}
	}
	else{
		switch (PlotType)
		{
			case 0:
				p.Create(NumPlotPoints,4);
				if(ProblemType==0){
					strcpy(p.lbls[1],"|A|, Wb/m");
					strcpy(p.lbls[2],"Re[A], Wb/m");
					strcpy(p.lbls[3],"Im[A], Wb/m");
				}
				else{
					strcpy(p.lbls[1],"|Flux|, Wb");
					strcpy(p.lbls[2],"Re[Flux], Wb");
					strcpy(p.lbls[3],"Im[Flux], Wb");
				}
				break;
			case 1:
				p.Create(NumPlotPoints,2);
				strcpy(p.lbls[1],"|B|, Tesla");
				break;
			case 2:
				p.Create(NumPlotPoints,4);
				strcpy(p.lbls[1],"|B.n|, Tesla");
				strcpy(p.lbls[2],"Re[B.n], Tesla");
				strcpy(p.lbls[3],"Im[B.n], Tesla");
				break;
			case 3:
				p.Create(NumPlotPoints,4);
				strcpy(p.lbls[1],"|B.t|, Tesla");
				strcpy(p.lbls[2],"Re[B.t], Tesla");
				strcpy(p.lbls[3],"Im[B.t], Tesla");
				break;
			case 4:
				p.Create(NumPlotPoints,2);
				strcpy(p.lbls[1],"|H|, Amp/m");
				break;
			case 5:
				p.Create(NumPlotPoints,4);
				if(ProblemType==0){
					strcpy(p.lbls[1],"|H.n|, Amp/m");
					strcpy(p.lbls[2],"Re[H.n], Amp/m");
					strcpy(p.lbls[3],"Im[H.n], Amp/m");
				}
				break;
			case 6:
				p.Create(NumPlotPoints,4);
				strcpy(p.lbls[1],"|H.t|, Amp/m");
				strcpy(p.lbls[2],"Re[H.t], Amp/m");
				strcpy(p.lbls[3],"Im[H.t], Amp/m");
				break;
			case 7:
				p.Create(NumPlotPoints,4);
				strcpy(p.lbls[1],"|Je|, MA/m^2");
				strcpy(p.lbls[2],"Re[Je], MA/m^2");
				strcpy(p.lbls[3],"Im[Je], MA/m^2");
				break;
			case 8:
				p.Create(NumPlotPoints,4);
				strcpy(p.lbls[1],"|Js+Je|, MA/m^2");
				strcpy(p.lbls[2],"Re[Js+Je], MA/m^2");
				strcpy(p.lbls[3],"Im[Js+Je], MA/m^2");
				break;
			default:
				p.Create(NumPlotPoints,2);
				break;
		}
	}
	
	switch(LengthUnits){
		case 1:
			strcpy(p.lbls[0],"Length, mm");
			break;
		case 2:
			strcpy(p.lbls[0],"Length, cm");
			break;
		case 3:
			strcpy(p.lbls[0],"Length, m");
			break;
		case 4:
			strcpy(p.lbls[0],"Length, mils");
			break;
		case 5:
			strcpy(p.lbls[0],"Length, um");
			break;
		default:
			strcpy(p.lbls[0],"Length, inches");
			break;
	}

	for(i=0,k=1,z=0,elm=-1;i<NumPlotPoints;i++,z+=dz)
	{
		while((z>q[k]) && (k<(contour.GetSize()-1))) k++;
		u=(z-q[k-1])/(q[k]-q[k-1]);
		pt=contour[k-1]+u*(contour[k]-contour[k-1]);
		t=contour[k]-contour[k-1];
		t/=abs(t);
		n = I*t;
		pt+=(n*1.e-06);
		
		if (elm<0) elm=InTriangle(pt.re,pt.im);
		else if (InTriangleTest(pt.re,pt.im,elm)==FALSE)
		{
			flag=FALSE;
			for(j=0;j<3;j++)
				for(m=0;m<NumList[meshelem[elm].p[j]];m++)
				{
					elm=ConList[meshelem[elm].p[j]][m];
					if (InTriangleTest(pt.re,pt.im,elm)==TRUE)
					{
						flag=TRUE;
						m=100;
						j=3;
					}
				}
			if (flag==FALSE) elm=InTriangle(pt.re,pt.im);
		}
		if(elm>=0)
			flag=GetPointValues(pt.re,pt.im,elm,v);
		else flag=FALSE;

		p.M[i][0]=z;
		if ((Frequency==0) && (flag!=FALSE)){
			switch (PlotType){
				case 0:
					p.M[i][1]=v.A.re;
					break;
				case 1:
					p.M[i][1]=sqrt(v.B1.Abs()*v.B1.Abs() + v.B2.Abs()*v.B2.Abs());
					break;
				case 2:
					p.M[i][1]= n.re*v.B1.re + n.im*v.B2.re;
					break;
				case 3:
					p.M[i][1]= t.re*v.B1.re + t.im*v.B2.re;
					break;
				case 4:
					p.M[i][1]=sqrt(v.H1.Abs()*v.H1.Abs() + v.H2.Abs()*v.H2.Abs());
					break;
				case 5:
					p.M[i][1]= n.re*v.H1.re + n.im*v.H2.re;
					break;
				case 6:
					p.M[i][1]= t.re*v.H1.re + t.im*v.H2.re;
					break;
				default:
					p.M[i][1]=0;
					break;
			}
		}
		else if (flag!=FALSE){
			switch (PlotType){
				case 0:
					p.M[i][1]=v.A.Abs();
					p.M[i][2]=v.A.re;
					p.M[i][3]=v.A.im;
					break;
				case 1:
					p.M[i][1]=sqrt(v.B1.Abs()*v.B1.Abs() + v.B2.Abs()*v.B2.Abs());
					break;
				case 2:
					p.M[i][2]= n.re*v.B1.re + n.im*v.B2.re;
					p.M[i][3]= n.re*v.B1.im + n.im*v.B2.im;
					p.M[i][1]=sqrt(p.M[i][2]*p.M[i][2] +
						           p.M[i][3]*p.M[i][3]);
					break;
				case 3:
					p.M[i][2]= t.re*v.B1.re + t.im*v.B2.re;
					p.M[i][3]= t.re*v.B1.im + t.im*v.B2.im;
					p.M[i][1]=sqrt(p.M[i][2]*p.M[i][2] +
						           p.M[i][3]*p.M[i][3]);
					break;
				case 4:
					p.M[i][1]=sqrt(v.H1.Abs()*v.H1.Abs() + v.H2.Abs()*v.H2.Abs());
					break;
				case 5:
					p.M[i][2]= n.re*v.H1.re + n.im*v.H2.re;
					p.M[i][3]= n.re*v.H1.im + n.im*v.H2.im;
					p.M[i][1]=sqrt(p.M[i][2]*p.M[i][2] +
						           p.M[i][3]*p.M[i][3]);
					break;
				case 6:
					p.M[i][2]= t.re*v.H1.re + t.im*v.H2.re;
					p.M[i][3]= t.re*v.H1.im + t.im*v.H2.im;
					p.M[i][1]=sqrt(p.M[i][2]*p.M[i][2] +
						           p.M[i][3]*p.M[i][3]);
					break;
				case 7:
					p.M[i][2]= v.Je.re;
					p.M[i][3]= v.Je.im;
					p.M[i][1]= abs(v.Je);
					break;
				case 8:
					p.M[i][2]= v.Je.re+v.Js.re;
					p.M[i][3]= v.Je.im+v.Js.im;
					p.M[i][1]= abs(Add(v.Je,v.Js));
					break;
				default:
					p.M[i][1]=0;
					break;
			}
		}
	}
		
	free(q);
}

BOOL CFemmviewDoc::InTriangleTest(double x, double y, int i)
{
	int j,k;
	double z;
	BOOL InFlag;

	if(i<0) return FALSE;

	for(j=0,InFlag=TRUE;((j<3) && (InFlag==TRUE));j++)
	{
		k=j+1; if(k==3) k=0;
		z=(meshnode[meshelem[i].p[k]].x-meshnode[meshelem[i].p[j]].x)*
		  (y-meshnode[meshelem[i].p[j]].y) -
		  (meshnode[meshelem[i].p[k]].y-meshnode[meshelem[i].p[j]].y)*
		  (x-meshnode[meshelem[i].p[j]].x);
		if(z<0) InFlag=FALSE;
	}

	return InFlag;
}

CPointVals::CPointVals()
{
	A.Set(0,0);			// vector potential
	B1.Set(0,0); B2.Set(0,0);		// flux density
	mu1.Set(1,0); mu2.Set(1,0);	// permeability
	H1.Set(0,0); H2.Set(0,0);		// field intensity
	Je.Set(0,0); Js.Set(0,0);		// eddy current and source current densities
	c=0;			// conductivity
	E=0;			// energy stored in the magnetic field
	Ph=0;			// power dissipated by hysteresis
	Pe=0;	
}

CComplex CFemmviewDoc::Ctr(int i)
{
	CComplex p,c;
	int j;
	
	for(j=0,c=0;j<3;j++){
		p.Set(meshnode[meshelem[i].p[j]].x/3.,meshnode[meshelem[i].p[j]].y/3.);
		c+=p;
	}

	return c;
}

double CFemmviewDoc::ElmArea(int i)
{
	int j,n[3];
	double b0,b1,c0,c1;

	for(j=0;j<3;j++) n[j]=meshelem[i].p[j];

	b0=meshnode[n[1]].y - meshnode[n[2]].y;
	b1=meshnode[n[2]].y - meshnode[n[0]].y;
	c0=meshnode[n[2]].x - meshnode[n[1]].x;
	c1=meshnode[n[0]].x - meshnode[n[2]].x;
	return (b0*c1-b1*c0)/2.;

}

double CFemmviewDoc::ElmArea(CElement *elm)
{
	int j,n[3];
	double b0,b1,c0,c1;

	for(j=0;j<3;j++) n[j]=elm->p[j];

	b0=meshnode[n[1]].y - meshnode[n[2]].y;
	b1=meshnode[n[2]].y - meshnode[n[0]].y;
	c0=meshnode[n[2]].x - meshnode[n[1]].x;
	c1=meshnode[n[0]].x - meshnode[n[2]].x;
	return (b0*c1-b1*c0)/2.;

}

void CFemmviewDoc::GetJA(int k,CComplex *J,CComplex *A)
{
	// returns current density with contribution from all sources in
	// units of MA/m^2

	int i,blk,lbl,crc;
	double r,c,rn;

	blk=meshelem[k].blk;
	lbl=meshelem[k].lbl;
	crc=blocklist[lbl].InCircuit;

	// first, get A
	for(i=0;i<3;i++){
		if(ProblemType==0) A[i]=(meshnode[meshelem[k].p[i]].A);
		else{
			rn=meshnode[meshelem[k].p[i]].x*LengthConv[LengthUnits];
			if(fabs(rn)<1.e-06) A[i]=0;
			else A[i]=(meshnode[meshelem[k].p[i]].A)/(2.*PI*rn);
		}
	}

	if(ProblemType==1) r=cRe(Ctr(k))*LengthConv[LengthUnits];

	// contribution from explicitly specified J
	for(i=0;i<3;i++) J[i]=blockproplist[blk].Jr+I*blockproplist[blk].Ji;

	if ((blockproplist[blk].Lam_d!=0) && (blockproplist[blk].LamType==0)) c=0;
	else c=blockproplist[blk].Cduct;

	// contribution from eddy currents;
	if(Frequency>0) for(i=0;i<3;i++) J[i]-=I*Frequency*2.*PI*c*A[i];

	// contribution from circuit currents //
	if(crc>=0)
	{
		if(circproplist[crc].Case==0){ // specified voltage
			if(ProblemType==0){
				for(i=0;i<3;i++) J[i]-=c*circproplist[crc].dVolts;
			}
			else{
				for(i=0;i<3;i++) J[i]-=c*circproplist[crc].dVolts/r;
			}
		}
		else 
			for(i=0;i<3;i++) J[i]+=circproplist[crc].J; // specified current
		
	}
		
	// convert results to A/m^2
	for(i=0;i<3;i++) J[i]*=1.e06;

	return;
}

CComplex CFemmviewDoc::PlnInt(double a, CComplex *u, CComplex *v)
{
	int i;
	CComplex z[3],x;
	
	z[0]=2.*u[0]+u[1]+u[2];
	z[1]=u[0]+2.*u[1]+u[2];
	z[2]=u[1]+u[1]+2.*u[2];

	for(i=0,x=0;i<3;i++) x+=v[i]*z[i];
	return a*x/12.;
}

CComplex CFemmviewDoc::AxiInt(double a, CComplex *u, CComplex *v,double *r)
{
	int i;
	static CComplex M[3][3];
	CComplex x, z[3];

	M[0][0]=6.*r[0]+2.*r[1]+2.*r[2];
	M[0][1]=2.*r[0]+2.*r[1]+1.*r[2];
	M[0][2]=2.*r[0]+1.*r[1]+2.*r[2];
	M[1][1]=2.*r[0]+6.*r[1]+2.*r[2];
	M[1][2]=1.*r[0]+2.*r[1]+2.*r[2];
	M[2][2]=2.*r[0]+2.*r[1]+6.*r[2];
	M[1][0]=M[0][1];
	M[2][0]=M[0][2];
	M[2][1]=M[1][2];
	
	for(i=0;i<3;i++) z[i]=M[i][0]*u[0]+M[i][1]*u[1]+M[i][2]*u[2];
	for(i=0,x=0;i<3;i++) x+=v[i]*z[i];
	return PI*a*x/30.;
}

CComplex CFemmviewDoc::BlockIntegral(int inttype)
{
	int i,k;
	CComplex c,y,z,J,mu1,mu2,B1,B2,H1,H2;
	CComplex A[3],Jn[3],U[3],V[3];
	double a,sig,R;
	double r[3];

	z=0; for(i=0;i<3;i++) U[i]=1.;
	
	if(inttype==6) z= BlockIntegral(3) + BlockIntegral(4); //total losses
	else for(i=0;i<meshelem.GetSize();i++)
	{
	  if(blocklist[meshelem[i].lbl].IsSelected==TRUE)
	  {

		// compute some useful quantities employed by most integrals...
		GetJA(i,Jn,A);
		for(k=0,J=0;k<3;k++) J+=Jn[k]/3;
		a=ElmArea(i)*pow(LengthConv[LengthUnits],2.);
		if(ProblemType==1){
			for(k=0;k<3;k++)
				r[k]=meshnode[meshelem[i].p[k]].x*LengthConv[LengthUnits];
			R=(r[0]+r[1]+r[2])/3.;
		}

		// now, compute the desired integral;
		switch(inttype)
		{
			case 0: //  A.J
				for(k=0;k<3;k++) V[k]=Jn[k].Conj();
				if(ProblemType==0)
					y=PlnInt(a,A,V);
				else
					y=AxiInt(a,A,V,r);
				z+=y;
				
				break;

			case 11: // x (or r) direction Lorentz force, SS part.
				B2=meshelem[i].B2;
				y= -(B2.re*J.re + B2.im*J.im);
				if (ProblemType==1) y=0;
				if(Frequency!=0) y*=0.5;
				z+=(a*y);
				break;

			case 12: // y (or z) direction Lorentz force, SS part.
				for(k=0;k<3;k++) V[k]=cRe(meshelem[i].B1*Jn[k].Conj());
				if(ProblemType==0)
					y=PlnInt(a,U,V);
				else
					y=AxiInt(-a,U,V,r);
				if(Frequency!=0) y*=0.5;
				z+=y;

				break;

			case 13: // x (or r) direction Lorentz force, 2x part.
				if((Frequency!=0) && (ProblemType==0))
				{
					B2=meshelem[i].B2;
					y= -(B2.re*J.re - B2.im*J.im) - I*(B2.re*J.im+B2.im*J.re);
					z+=0.5*(a*y);
				}
				break;

			case 14: // y (or z) direction Lorentz force, 2x part.
				if (Frequency!=0)
				{
					B1=meshelem[i].B1;
					B2=meshelem[i].B2;
					y= (B1.re*J.re - B1.im*J.im) + I*(B1.re*J.im+B1.im*J.re);
					if(ProblemType==1) y=(-y*2.*PI*R);
					z+=(a*y)/2.;
				}
				break;

			case 16: // Lorentz Torque, 2x
				if ((Frequency!=0) && (ProblemType==0))
				{
					B1=meshelem[i].B1;
					c=Ctr(i)*LengthConv[LengthUnits];
					y= c.re*((B1.re*J.re - B1.im*J.im) + I*(B1.re*J.im+B1.im*J.re))
					  +c.im*((B2.re*J.re - B2.im*J.im) + I*(B2.re*J.im+B2.im*J.re));
					z+=0.5*(a*y);
				}
				break;

			case 15: // Lorentz Torque, SS part.
				if(ProblemType==0)
				{
					B1=meshelem[i].B1;
					B2=meshelem[i].B2;
					c=Ctr(i)*LengthConv[LengthUnits];
					y= c.im*(B2.re*J.re + B2.im*J.im) + c.re*(B1.re*J.re + B1.im*J.im);
					if(Frequency!=0) y*=0.5;
					z+=(a*y);
				}
				break;

			case 1: // integrate A over the element;
				if(ProblemType==1)
					y=AxiInt(a,U,A,r);
				else
					for(k=0,y=0;k<3;k++) y+=a*A[k]/3.;

				z+=y;
				break;

			case 2: // stored energy
				if(Frequency!=0){
					if(ProblemType==1) a*=(2.*PI*R);
					mu1=blockproplist[meshelem[i].blk].mu_fdx;
					mu1*= 4.*PI*1.e-07;
					mu2=blockproplist[meshelem[i].blk].mu_fdy;
					mu2*= 4.*PI*1.e-07;
					B1=meshelem[i].B1;
					B2=meshelem[i].B2;
					H1=B1/mu1;
					H2=B2/mu2;
					y=a*0.25*cRe(H1*B1.Conj()+H2*B2.Conj());
					z+=y;
				}
				else{
					if(ProblemType==1) a*=(2.*PI*R);
				//	mu1.Set(meshelem[i].mu1,0);
				//	mu1=Mult(4.*PI*1.e-07,mu1);
				//	mu2.Set(meshelem[i].mu2,0);
				//	mu2=Mult(4.*PI*1.e-07,mu2);
					B1=meshelem[i].B1;
					B2=meshelem[i].B2;
				//	H1=B1/mu1;
				//	H2=B2/mu2;
				//	y=a*0.5*(H1.re*B1.re + H2.re*B2.re);
					
					y=a*blockproplist[meshelem[i].blk].DoEnergy(B1.re,B2.re);
					z+=y;
				}
				break;

			case 3:  // Hysteresis & Laminated eddy current losses
				if(Frequency!=0){
					if(ProblemType==1) a*=(2.*PI*R);
					mu1=blockproplist[meshelem[i].blk].mu_fdx;
					mu1*= 4.*PI*1.e-07;
					mu2=blockproplist[meshelem[i].blk].mu_fdy;
					mu2*= 4.*PI*1.e-07;
					B1=meshelem[i].B1;
					B2=meshelem[i].B2;
					H1=B1/mu1;
					H2=B2/mu2;
					y=a*PI*Frequency*cIm(H1*B1.Conj() + H2*B2.Conj());
					z+=y;
				}
				break;
			
			case 4: // Resistive Losses
				sig=blockproplist[meshelem[i].blk].Cduct * 1.e06;
				if((blockproplist[meshelem[i].blk].Lam_d!=0) &&
					(blockproplist[meshelem[i].blk].LamType==0)) sig=0;
				if(sig!=0){
					for(k=0;k<3;k++) V[k]=Jn[k].Conj()/sig;	
					if (ProblemType==0) y=PlnInt(a,Jn,V);
					if(ProblemType==1) y=AxiInt(a,Jn,V,r);
					if(Frequency!=0) y/=2.;
					z+=y;
				}
				break;
			
			case 5: // cross-section area
				z+=a;
				break;
			
            case 10: // volume
				if(ProblemType==1) a*=(2.*PI*R);
				z+=a;
				break;

			case 7: // total current in block;
				z+=a*J;

				break;

			case 8: // integrate x or r part of b over the block
				if(ProblemType==1) a*=(2.*PI*R);
				z+=(a*meshelem[i].B1);
				break;
			
			case 9: // integrate y or z part of b over the block
				if(ProblemType==1) a*=(2.*PI*R);
				z+=(a*meshelem[i].B2);
				break;

			case 17: // Coenergy
				if(Frequency!=0){
					// for harmonic problem, everything
					// is linear--just evaluate stored energy
					// per cycle
					if(ProblemType==1) a*=(2.*PI*R);
					mu1=blockproplist[meshelem[i].blk].mu_fdx;
					mu1*= 4.*PI*1.e-07;
					mu2=blockproplist[meshelem[i].blk].mu_fdy;
					mu2*= 4.*PI*1.e-07;
					B1=meshelem[i].B1;
					B2=meshelem[i].B2;
					H1=B1/mu1;
					H2=B2/mu2;
					y=a*0.25*cRe(H1*B1.Conj()+H2*B2.Conj());
					z+=y;
				}
				else{
					if(ProblemType==1) a*=(2.*PI*R);
					B1=meshelem[i].B1;
					B2=meshelem[i].B2;		
					y=a*blockproplist[meshelem[i].blk].DoCoEnergy(B1.re,B2.re);
					z+=y;
				}
				break;

			default:
				break;
		}
	
      }
	}

	return z;
}

void CFemmviewDoc::LineIntegral(int inttype, CComplex *z)
{
	if(inttype==0){
		CComplex a0,a1;
		CPointVals u;
		double l;
		int i,k;

		k=contour.GetSize();
		GetPointValues(contour[0].re,contour[0].im, u);
		a0=u.A;
		GetPointValues(contour[k-1].re,contour[k-1].im,u);
		a1=u.A;
		if(ProblemType==0){
			for(i=0,l=0;i<k-1;i++)
				l+=abs(contour[i+1]-contour[i]);
			l*=LengthConv[LengthUnits];
			z[0] = a0-a1;
			if(l!=0) z[1]= z[0]/l;
		}
		else{
			for(i=0,l=0;i<k-1;i++)
				l+=(PI*(contour[i].re+contour[i+1].re)*
					abs(contour[i+1]-contour[i]));
			l*=pow(LengthConv[LengthUnits],2.);
			z[0]= a1-a0;
			if(l!=0) z[1]= z[0]/l;
		}
	}

	if(inttype==1){
		CComplex n,t,pt,Ht;
		CPointVals v;
		double dz,u,l;
		int i,j,k,m,elm;
		int NumPlotPoints=400;
		BOOL flag;

		z[0]=0;
	for(k=1;k<contour.GetSize();k++)
	{
		dz=abs(contour[k]-contour[k-1])/((double) NumPlotPoints);
		for(i=0,elm=-1;i<NumPlotPoints;i++)
		{
			u=(((double) i)+0.5)/((double) NumPlotPoints);
			pt=Add(contour[k-1],Mult(u,Subt(contour[k],contour[k-1])));
			t=contour[k]-contour[k-1];
			t/=abs(t);
			n=I*t;
			pt+=n*1.e-06;
		
			if (elm<0) elm=InTriangle(pt.re,pt.im);
			else if (InTriangleTest(pt.re,pt.im,elm)==FALSE)
			{
				flag=FALSE;
				for(j=0;j<3;j++)
					for(m=0;m<NumList[meshelem[elm].p[j]];m++)
					{
						elm=ConList[meshelem[elm].p[j]][m];
						if (InTriangleTest(pt.re,pt.im,elm)==TRUE)
						{
							flag=TRUE;
							m=100;
							j=3;
						}
					}
				if (flag==FALSE) elm=InTriangle(pt.re,pt.im);
			}
			if(elm>=0)
				flag=GetPointValues(pt.re,pt.im,elm,v);
			else flag=FALSE;

			if(flag==TRUE){
				Ht = t.re*v.H1 + t.im*v.H2;
				z[0]+=(Ht*dz*LengthConv[LengthUnits]);
			}
		}


		for(i=0,l=0;i<contour.GetSize()-1;i++)
			l+=abs(contour[i+1]-contour[i]);
		l*=LengthConv[LengthUnits];
		if(l!=0) z[1]=z[0]/l;

	}
	}

	if(inttype==2){
		int i,k;
		k=contour.GetSize();
		for(i=0,z[0].re=0;i<k-1;i++)
			z[0].re+=abs(contour[i+1]-contour[i]);
		z[0].re*=LengthConv[LengthUnits];

		if(ProblemType==1){
			for(i=0,z[0].im=0;i<k-1;i++)
				z[0].im+=(PI*(contour[i].re+contour[i+1].re)*
						abs(contour[i+1]-contour[i]));
			z[0].im*=pow(LengthConv[LengthUnits],2.);
		}
	}

	if(inttype==3){
		CComplex n,t,pt,Hn,Bn,BH,dF1,dF2;
		CPointVals v;
		double dz,dza,u;
		int i,j,k,m,elm;
		int NumPlotPoints=400;
		BOOL flag;

		for(i=0;i<4;i++) z[i]=0;

	for(k=1;k<contour.GetSize();k++)
	{
		dz=abs(contour[k]-contour[k-1])/((double) NumPlotPoints);
		for(i=0,elm=-1;i<NumPlotPoints;i++)
		{
			u=(((double) i)+0.5)/((double) NumPlotPoints);
			pt=Add(contour[k-1],Mult(u,Subt(contour[k],contour[k-1])));
			t=contour[k]-contour[k-1];
			t/=abs(t);
			n=I*t;
			pt+=n*1.e-06;
		
			if (elm<0) elm=InTriangle(pt.re,pt.im);
			else if (InTriangleTest(pt.re,pt.im,elm)==FALSE)
			{
				flag=FALSE;
				for(j=0;j<3;j++)
					for(m=0;m<NumList[meshelem[elm].p[j]];m++)
					{
						elm=ConList[meshelem[elm].p[j]][m];
						if (InTriangleTest(pt.re,pt.im,elm)==TRUE)
						{
							flag=TRUE;
							m=100;
							j=3;
						}
					}
				if (flag==FALSE) elm=InTriangle(pt.re,pt.im);
			}
			if(elm>=0)
				flag=GetPointValues(pt.re,pt.im,elm,v);
			else flag=FALSE;

			if(flag==TRUE)
			if(Frequency==0){
				Hn= n.re*v.H1 + n.im*v.H2;
				Bn= n.re*v.B1 + n.im*v.B2;
				BH= v.B1*v.H1 + v.B2*v.H2;
				dF1=v.H1*Bn + v.B1*Hn - n.re*BH;
				dF2=v.H2*Bn + v.B2*Hn - n.im*BH;
				
				dza=dz*LengthConv[LengthUnits];
				if(ProblemType==1) dza*=2.*PI*pt.re*LengthConv[LengthUnits];
				
				z[0]+=(dF1*dza/2.);
				z[1]+=(dF2*dza/2.);
			}
			else{
				Hn=n.re*v.H1 + n.im*v.H2;
				Bn=n.re*v.B1 + n.im*v.B2;
				BH = v.B1*v.H1 + v.B2*v.H2;
				dF1 = v.H1*Bn + v.B1*Hn - n.re*BH;
				dF2 = v.H2*Bn + v.B2*Hn - n.im*BH;
				
				dza=dz*LengthConv[LengthUnits];
				if(ProblemType==1) dza*=2.*PI*pt.re*LengthConv[LengthUnits];
				
				z[0]+=(dF1*dza/4.);
				z[1]+=(dF2*dza/4.);
			
				BH  = v.B1*v.H1.Conj() +v.B2*v.H2.Conj();
				dF1 = v.H1*Bn.Conj() + v.B1*Hn.Conj() - n.re*BH;
				dF2=  v.H2*Bn.Conj() + v.B2*Hn.Conj() - n.im*BH;

				z[2]+=(dF1*dza/4.);
				z[3]+=(dF2*dza/4.);
			}
		}
		
	}
	}

	if(inttype==4){
		CComplex n,t,pt,Hn,Bn,BH,dF1,dF2,dT;
		CPointVals v;
		double dz,dza,u;
		int i,j,k,m,elm;
		int NumPlotPoints=400;
		BOOL flag;

		for(i=0;i<2;i++) z[i].Set(0,0);

	for(k=1;k<contour.GetSize();k++)
	{
		dz=abs(contour[k]-contour[k-1])/((double) NumPlotPoints);
		for(i=0,elm=-1;i<NumPlotPoints;i++)
		{
			u=(((double) i)+0.5)/((double) NumPlotPoints);
			pt=Add(contour[k-1],Mult(u,Subt(contour[k],contour[k-1])));
			t=Subt(contour[k],contour[k-1]);
			t=Div(t,abs(t));
			n.Set(-t.im,t.re);
			pt=Add(pt,Mult(1.e-06,n));
		
			if (elm<0) elm=InTriangle(pt.re,pt.im);
			else if (InTriangleTest(pt.re,pt.im,elm)==FALSE)
			{
				flag=FALSE;
				for(j=0;j<3;j++)
					for(m=0;m<NumList[meshelem[elm].p[j]];m++)
					{
						elm=ConList[meshelem[elm].p[j]][m];
						if (InTriangleTest(pt.re,pt.im,elm)==TRUE)
						{
							flag=TRUE;
							m=100;
							j=3;
						}
					}
				if (flag==FALSE) elm=InTriangle(pt.re,pt.im);
			}
			if(elm>=0)
				flag=GetPointValues(pt.re,pt.im,elm,v);
			else flag=FALSE;

			if(flag==TRUE)
			if(Frequency==0){
				Hn= n.re*v.H1 + n.im*v.H2;
				Bn= n.re*v.B1 + n.im*v.B2;
				BH= v.B1*v.H1 + v.B2*v.H2;
				dF1=v.H1*Bn + v.B1*Hn - n.re*BH;
				dF2=v.H2*Bn + v.B2*Hn - n.im*BH;
				dT= pt.re*dF2 - dF1*pt.im;				
				dza=dz*LengthConv[LengthUnits]*LengthConv[LengthUnits];
						
				z[0]+=(dT*dza/2.);
			}
			else{
				Hn=n.re*v.H1 + n.im*v.H2;
				Bn=n.re*v.B1 + n.im*v.B2;
				BH = v.B1*v.H1 + v.B2*v.H2;
				dF1 = v.H1*Bn + v.B1*Hn - n.re*BH;
				dF2 = v.H2*Bn + v.B2*Hn - n.im*BH;
				dT=pt.re*dF2 - dF1*pt.im;	
				dza=dz*LengthConv[LengthUnits]*LengthConv[LengthUnits];
								
				z[0]+=(dT*dza/4.);
			
				BH  = v.B1*v.H1.Conj() +v.B2*v.H2.Conj();
				dF1 = v.H1*Bn.Conj() + v.B1*Hn.Conj() - n.re*BH;
				dF2=  v.H2*Bn.Conj() + v.B2*Hn.Conj() - n.im*BH;
				dT= pt.re*dF2 - dF1*pt.im ;	

				z[1]+=(dT*dza/4.);
			
			}
		}
	}
	
	}

	if(inttype==5){
		CComplex n,t,pt,Ht;
		CPointVals v;
		double dz,u,l;
		int i,j,k,m,elm;
		int NumPlotPoints=400;
		BOOL flag;

		z[0]=0;
	for(k=1;k<contour.GetSize();k++)
	{
		dz=abs(contour[k]-contour[k-1])/((double) NumPlotPoints);
		for(i=0,elm=-1;i<NumPlotPoints;i++)
		{
			u=(((double) i)+0.5)/((double) NumPlotPoints);
			pt=Add(contour[k-1],Mult(u,Subt(contour[k],contour[k-1])));
			t=contour[k]-contour[k-1];
			t/=abs(t);
			n=I*t;
			pt+=n*1.e-06;
		
			if (elm<0) elm=InTriangle(pt.re,pt.im);
			else if (InTriangleTest(pt.re,pt.im,elm)==FALSE)
			{
				flag=FALSE;
				for(j=0;j<3;j++)
					for(m=0;m<NumList[meshelem[elm].p[j]];m++)
					{
						elm=ConList[meshelem[elm].p[j]][m];
						if (InTriangleTest(pt.re,pt.im,elm)==TRUE)
						{
							flag=TRUE;
							m=100;
							j=3;
						}
					}
				if (flag==FALSE) elm=InTriangle(pt.re,pt.im);
			}
			if(elm>=0)
				flag=GetPointValues(pt.re,pt.im,elm,v);
			else flag=FALSE;

			if(flag==TRUE){
				Ht = n.re*v.B1 + n.im*v.B2;
				z[0]+=(Ht*Ht.Conj()*dz*LengthConv[LengthUnits]);
			}
		}


		for(i=0,l=0;i<contour.GetSize()-1;i++)
			l+=abs(contour[i+1]-contour[i]);
		l*=LengthConv[LengthUnits];
		if(l!=0) z[1]=z[0]/l;

	}
	}

	return;
}

int CFemmviewDoc::ClosestArcSegment(double x, double y)
{
	double d0,d1;
	int i,j;

	if(arclist.GetSize()==0) return -1;

	j=0;
	d0=ShortestDistanceFromArc(CComplex(x,y),arclist[0]);
	for(i=0;i<arclist.GetSize();i++){
		d1=ShortestDistanceFromArc(CComplex(x,y),arclist[i]);
		if(d1<d0){
			d0=d1;
			j=i;
		}
	}

	return j;
}

void CFemmviewDoc::GetCircle(CArcSegment &arc,CComplex &c, double &R)
{
	CComplex a0,a1,t;
	double d,tta;

	a0.Set(nodelist[arc.n0].x,nodelist[arc.n0].y);
	a1.Set(nodelist[arc.n1].x,nodelist[arc.n1].y);
	d=abs(a1-a0);			// distance between arc endpoints

	// figure out what the radius of the circle is...
	t=(a1-a0)/d;
	tta=arc.ArcLength*PI/180.;
	R=d/(2.*sin(tta/2.));
	c=a0 + (d/2. + I*sqrt(R*R-d*d/4.))*t; // center of the arc segment's circle...
}

double CFemmviewDoc::ShortestDistanceFromArc(CComplex p, CArcSegment &arc)
{
	double R,d,l,z;
	CComplex a0,a1,c,t;

	a0.Set(nodelist[arc.n0].x,nodelist[arc.n0].y);
	a1.Set(nodelist[arc.n1].x,nodelist[arc.n1].y);
	GetCircle(arc,c,R);
	d=abs(p-c);
	
	if(d==0) return R;
	
	t=(p-c)/d;
	l=abs(p-c-R*t);
	z=arg(t/(a0-c))*180/PI;
	if ((z>0) && (z<arc.ArcLength)) return l;
	
	z=abs(p-a0);
	l=abs(p-a1);
	if(z<l) return z;
	return l;
}

double CFemmviewDoc::ShortestDistanceFromSegment(double p, double q, int segm)
{
	double t,x[3],y[3];
	
	x[0]=nodelist[linelist[segm].n0].x;
	y[0]=nodelist[linelist[segm].n0].y;
	x[1]=nodelist[linelist[segm].n1].x;
	y[1]=nodelist[linelist[segm].n1].y;

	t=((p-x[0])*(x[1]-x[0]) + (q-y[0])*(y[1]-y[0]))/
	  ((x[1]-x[0])*(x[1]-x[0]) + (y[1]-y[0])*(y[1]-y[0]));

	if (t>1.) t=1.;
	if (t<0.) t=0.;

	x[2]=x[0]+t*(x[1]-x[0]);
	y[2]=y[0]+t*(y[1]-y[0]);

	return sqrt((p-x[2])*(p-x[2]) + (q-y[2])*(q-y[2]));
}

