// Solve linear 2-D problems...
#include<stdafx.h>
#include<stdio.h>
#include<math.h>
#include "fkn.h"
#include "fknDlg.h"
#include "complex.h"
#include "mesh.h"
#include "spars.h"
#include "FemmeDocCore.h"

BOOL CFemmeDocCore::Harmonic2D(CBigComplexLinProb &L)
{
	int i,j,k,s,sdin,sdi_iter,pctr;
	CComplex Me[3][3],be[3];		// element matrices;
	double l[3],p[3],q[3];		// element shape parameters;
	int n[3];					// numbers of nodes for a particular element;
	double a,r,t,x,y,ds,w;
	CComplex K,mu1,mu2,lag,halflag,deg45,Jv;
	CComplex **Mu,*V_sdi;
	double c=PI*4.e-05;
	double units[]={2.54,0.1,1.,100.,0.00254,1.e-04};
	CElement *El;
	deg45=1+I;
	w=Frequency*2.*PI;
	BOOL SDIflag=FALSE;
	
	CComplex *CircInt1,*CircInt2,*CircInt3;
	
	// check to see if any circuits have been defined and process them;
	if (NumCircProps>0)
	{
		CircInt1=(CComplex *)calloc(NumCircProps,sizeof(CComplex));
		CircInt2=(CComplex *)calloc(NumCircProps,sizeof(CComplex));
		CircInt3=(CComplex *)calloc(NumCircProps,sizeof(CComplex));
		for(i=0;i<NumEls;i++){
			if(meshele[i].lbl>=0)
			if(labellist[meshele[i].lbl].InCircuit!=-1){
				El=&meshele[i];
				
				// get element area;
				for(k=0;k<3;k++) n[k]=El->p[k];
				p[0]=meshnode[n[1]].y - meshnode[n[2]].y;
				p[1]=meshnode[n[2]].y - meshnode[n[0]].y;
				p[2]=meshnode[n[0]].y - meshnode[n[1]].y;	
				q[0]=meshnode[n[2]].x - meshnode[n[1]].x;
				q[1]=meshnode[n[0]].x - meshnode[n[2]].x;
				q[2]=meshnode[n[1]].x - meshnode[n[0]].x;
				a=(p[0]*q[1]-p[1]*q[0])/2.;
			//	r=(meshnode[n[0]].x+meshnode[n[1]].x+meshnode[n[2]].x)/3.;

			// evaluate integrals;
				
				// total cross-section of circuit;
				CircInt1[labellist[El->lbl].InCircuit]+=a; 

				// integral of conductivity over the circuit;
				CircInt2[labellist[El->lbl].InCircuit]+=
					a*blockproplist[El->blk].Cduct;

				// integral of applied J over current;
				CircInt3[labellist[El->lbl].InCircuit]+=
					(blockproplist[El->blk].Jr+I*blockproplist[El->blk].Ji)*a*100.;
			}
		}

		for(i=0;i<NumCircProps;i++)
		{
			if (circproplist[i].CircType==0) // specified current
			{
				if(CircInt2[i]==0){ //circuit composed of zero cond. materials
					circproplist[i].Case=1;
					if (CircInt1[i]==0.) circproplist[i].J=0.;
					else circproplist[i].J=0.01*(
						(circproplist[i].Amps_re+I*circproplist[i].Amps_im) -
						 CircInt3[i])/CircInt1[i];
				}
				else{
					circproplist[i].Case=2; // need to include an extra
										    // entry in matrix to solve for
					                        // voltage grad in the circuit
				}
			}
			else{
				// case where voltage gradient is specified a priori...
				circproplist[i].Case=0;
				circproplist[i].dV=circproplist[i].dVolts_re +
								   I*circproplist[i].dVolts_im;
			}
		}
	}



	// check to see if there are any SDI boundaries...
	// lineproplist[ meshele[i].e[j] ].BdryFormat==0
	for(i=0;i<NumLineProps;i++)
		if(lineproplist[i].BdryFormat==3) SDIflag=TRUE;

	if(SDIflag==TRUE){
		// there is an SDI boundary defined; check to see if it is in use
		SDIflag=FALSE;
		for(i=0;i<NumEls;i++)
			for(j=0;j<3;j++)
				if (lineproplist[meshele[i].e[j]].BdryFormat==3){
					SDIflag=TRUE;
					printf("Problem has SDI boundaries\n");
					i=NumEls;
					j=3;
				}
	}

	if (SDIflag==TRUE)
	{
		V_sdi=(CComplex *) calloc(NumNodes+NumCircProps,sizeof(CComplex));
		sdin=2;
	}
	else sdin=1;

  for(sdi_iter=0; sdi_iter<sdin; sdi_iter++)
  {
		TheView->SetDlgItemText(IDC_FRAME1,"Matrix Construction");
		TheView->m_prog1.SetPos(0);
		pctr=0;

   if (sdi_iter>0) L.Wipe();
   else{
	// compute effective permeability for each block type;
	Mu=(CComplex **)calloc(NumBlockProps,sizeof(CComplex *));
	for(i=0;i<NumBlockProps;i++) Mu[i]=(CComplex *)calloc(2,sizeof(CComplex));

	for(k=0;k<NumBlockProps;k++){
		lag=exp(-I*blockproplist[k].Theta_h*PI/180.);
		if (blockproplist[k].LamType==0){
			Mu[k][0]=blockproplist[k].mu_x*lag;
			Mu[k][1]=blockproplist[k].mu_y*lag;
			
			if(blockproplist[k].Lam_d!=0){
				halflag=exp(-I*blockproplist[k].Theta_h*PI/360.);

				ds=sqrt(2./(0.4*PI*w*blockproplist[k].Cduct*blockproplist[k].mu_x));
				K=halflag*deg45*blockproplist[k].Lam_d*0.001/(2.*ds);
				Mu[k][0]=((Mu[k][0]*tanh(K))/K)*blockproplist[k].LamFill;
	
				ds=sqrt(2./(0.4*PI*w*blockproplist[k].Cduct*blockproplist[k].mu_y));
				K=halflag*deg45*blockproplist[k].Lam_d*0.001/(2.*ds);
				Mu[k][1]=((Mu[k][1]*tanh(K))/K)*blockproplist[k].LamFill;
			}
		}
		if (blockproplist[k].LamType==1){
				x=blockproplist[k].LamFill;
				K=lag*blockproplist[k].mu_x;
				Mu[k][0]=(K*x) + (1.-x);
				Mu[k][1]=K/(x + K*(1.-x));
		}

		if (blockproplist[k].LamType==2){
				x=blockproplist[k].LamFill;
				K=lag*blockproplist[k].mu_y;
				Mu[k][1]=(K*x) + (1.-x);
				Mu[k][0]=K/(x + K*(1.-x));
		}
	}
   }

	// build element matrices using the matrices derived in Allaire's book.
	for(i=0;i<NumEls;i++){
		
		// update ``building matrix'' progress bar...
		j=(i*20)/NumEls+1;
		if(j>pctr){
			j=pctr*5; if (j>100) j=100;
			TheView->m_prog1.SetPos(j);
			pctr++;
		}
		
		// zero out Me, be;
		for(j=0;j<3;j++){
			for(k=0;k<3;k++) Me[j][k]=0;
			be[j]=0;
		}

		// Determine shape parameters.
		// l == element side lengths;
		// p corresponds to the `b' parameter in Allaire
		// q corresponds to the `c' parameter in Allaire
		El=&meshele[i];		
	
		for(k=0;k<3;k++) n[k]=El->p[k];
		p[0]=meshnode[n[1]].y - meshnode[n[2]].y;
		p[1]=meshnode[n[2]].y - meshnode[n[0]].y;
		p[2]=meshnode[n[0]].y - meshnode[n[1]].y;	
		q[0]=meshnode[n[2]].x - meshnode[n[1]].x;
		q[1]=meshnode[n[0]].x - meshnode[n[2]].x;
		q[2]=meshnode[n[1]].x - meshnode[n[0]].x;
		for(j=0,k=1;j<3;k++,j++){
			if (k==3) k=0;
			l[j]=sqrt( pow(meshnode[n[k]].x-meshnode[n[j]].x,2.) +
					   pow(meshnode[n[k]].y-meshnode[n[j]].y,2.) );
		}
		a=(p[0]*q[1]-p[1]*q[0])/2.;

		// retrieve permeability for the element;
		k=meshele[i].blk;
		mu1=Mu[k][0];
		mu2=Mu[k][1];	

		// x-contribution; only need to do main diagonal and above;
		K = -1./(4.*a*mu2);

		for(j=0;j<3;j++)
			for(k=j;k<3;k++)
				Me[j][k]+=(K*p[j]*p[k]);
		
		// y-contribution; only need to do main diagonal and above;
		K = -1./(4.*a*mu1);
		
		for(j=0;j<3;j++)
			for(k=j;k<3;k++)
				Me[j][k]+=(K*q[j]*q[k]);

		// contribution from eddy currents;
		
		K=-I*a*w*blockproplist[meshele[i].blk].Cduct*c/12.;
		if((blockproplist[El->blk].LamType==0) &&
			(blockproplist[El->blk].Lam_d>0)) K.Set(0,0);
			// because in-plane laminated blocks appear to have no conductivity;
			// eddy currents are accounted for in these elements by their
			// frequency-dependent permeability.
		for(j=0;j<3;j++)
			for(k=j;k<3;k++){
				Me[j][k]+=K;
				if(j==k) Me[j][k]+=K;
			}
	
		// contributions to Me, be from derivative boundary conditions;
		for(j=0;j<3;j++){
			if (El->e[j] >= 0)
			if (lineproplist[El->e[j]].BdryFormat==2)
			{
				// conversion factor is 10^(-4) (I think...)
				K.Set(-0.0001*c*lineproplist[ El->e[j] ].c0*l[j]/6.,0);
				k=j+1; if(k==3) k=0;
				Me[j][j]+=2*K;
				Me[k][k]+=2*K;
				Me[j][k]+=K;
				Me[k][j]+=K;

				K=(lineproplist[ El->e[j] ].c1*l[j]/2.)*0.0001;
				be[j]+=K;
				be[k]+=K;
			}
			
			if (lineproplist[El->e[j]].BdryFormat==1)
			{
				ds=sqrt(2./(0.4*PI*w*lineproplist[El->e[j]].Sig*
					lineproplist[El->e[j]].Mu));
				K=deg45/(-ds*lineproplist[El->e[j]].Mu*100.);
				K*=(l[j]/6.);
				k=j+1; if(k==3) k=0;
				Me[j][j]+=2*K;
				Me[k][k]+=2*K;
				Me[j][k]+=K;
				Me[k][j]+=K;
			}
		}
		
/*		// contribution to be from current density in the block
		// (version without modifications for circuits
		for(j=0;j<3;j++){
			K.Set(blockproplist[El->blk].Jr*a/3.,blockproplist[El->blk].Ji*a/3.);
			be[j]+=(-K);
		}	*/

		// contribution to be from current density in the block
		for(j=0;j<3;j++){
			Jv=0;
			if(labellist[El->lbl].InCircuit>=0)
			{
				k=labellist[El->lbl].InCircuit;
				if(circproplist[k].Case==1) Jv=circproplist[k].J;
				if(circproplist[k].Case==0)
					Jv=-circproplist[k].dV*blockproplist[El->blk].Cduct;
			}
			K=-(blockproplist[El->blk].Jr+I*blockproplist[El->blk].Ji+Jv)*a/3.;
			be[j]+=K;

			if(labellist[El->lbl].InCircuit>=0){
				k=labellist[El->lbl].InCircuit;
				if(circproplist[k].Case==2) L.b[NumNodes+k]+=K;
			}
		}

		// do Case 2 circuit stuff for element
		if(labellist[El->lbl].InCircuit>=0){
			k=labellist[El->lbl].InCircuit;
			if(circproplist[k].Case==2){
				K=-I*a*w*blockproplist[meshele[i].blk].Cduct*c;
				for(j=0;j<3;j++) L.Put(L.Get(n[j],NumNodes+k)+K/3.,n[j],NumNodes+k);
				L.Put(L.Get(NumNodes+k,NumNodes+k)+K,NumNodes+k,NumNodes+k);
			}
		}

/*
		// contribution to be from magnetization in the block;
		for(j=0;j<3;j++){
			t=blockproplist[El->blk].Theta_m;
			k=j+1; if(k==3) k=0;
			// need to scale so that everything is in proper units...
			// conversion is 0.0001
			K=0.0001*blockproplist[El->blk].H_c*(
			  cos(t*PI/180.)*(meshnode[n[k]].x-meshnode[n[j]].x) +
			  sin(t*PI/180.)*(meshnode[n[k]].y-meshnode[n[j]].y) )/2.;
			be[j]+=K;
			be[k]+=K;
		}	
*/

		// combine block matrices into global matrices;
		for (j=0;j<3;j++){
			for (k=j;k<3;k++)
				L.Put( L.Get(n[j],n[k]) + Me[j][k], n[j],n[k]);
			L.b[n[j]]+=be[j];
		}
	}

	// add in contribution from point currents;
	for(i=0;i<NumNodes;i++)
		if(meshnode[i].bc>=0){
			K.Set(0.01*nodeproplist[meshnode[i].bc].Jr,
				  0.01*nodeproplist[meshnode[i].bc].Ji);
			L.b[i]+=(-K);
		}

	// add in total current constraints for circuits;
	for(i=0;i<NumCircProps;i++)
		if (circproplist[i].Case==2){
			L.b[NumNodes+i]+=0.01*(circproplist[i].Amps_re +
				                   I*circproplist[i].Amps_im);
		}

	// apply fixed boundary conditions at points;
	for(i=0;i<NumNodes;i++)
		if(meshnode[i].bc >=0)
			if((nodeproplist[meshnode[i].bc].Jr==0) &&
			   (nodeproplist[meshnode[i].bc].Ji==0) && (sdi_iter==0))
			{
				K.Set(nodeproplist[meshnode[i].bc].Ar / c,
				  nodeproplist[meshnode[i].bc].Ai / c);
				L.SetValue(i,K);
			}

	// apply fixed boundary conditions along segments;
	for(i=0;i<NumEls;i++)
		for(j=0;j<3;j++){
			k=j+1; if(k==3) k=0;
			if(meshele[i].e[j]>=0)
			if(lineproplist[ meshele[i].e[j] ].BdryFormat==0)
			{
				if(Coords==0){
					// first point on the side;
					x=meshnode[meshele[i].p[j]].x;
					y=meshnode[meshele[i].p[j]].y;
					x*=units[LengthUnits]; y*=units[LengthUnits];
					s=meshele[i].e[j];
					a=lineproplist[s].A0 + x*lineproplist[s].A1 +
					  y*lineproplist[s].A2;
					// convert into appropriate imaginary number.
					K.Set(a*cos(lineproplist[s].phi*180./PI)/c,
						a*sin(lineproplist[s].phi*180./PI)/c);
					L.SetValue(meshele[i].p[j],K);
					
					// second point on the side;
					x=meshnode[meshele[i].p[k]].x;
					y=meshnode[meshele[i].p[k]].y;
					x*=units[LengthUnits]; y*=units[LengthUnits];
					s=meshele[i].e[j];
					a=lineproplist[s].A0 + x*lineproplist[s].A1 +
					  y*lineproplist[s].A2;
					// just take``real'' component.
					K.Set(a*cos(lineproplist[s].phi*180./PI)/c,
						a*sin(lineproplist[s].phi*180./PI)/c);
					L.SetValue(meshele[i].p[k],K);
				}
				else{
					// first point on the side;
					x=meshnode[meshele[i].p[j]].x;
					y=meshnode[meshele[i].p[j]].y;
					r=sqrt(x*x+y*y);
					if ((x==0) && (y==0)) t=0;
					else t=atan2(y,x)*180./PI;
					r*=units[LengthUnits];
					s=meshele[i].e[j];
					a=lineproplist[s].A0 + r*lineproplist[s].A1 +
					  t*lineproplist[s].A2;
					K.Set(a*cos(lineproplist[s].phi*180./PI)/c,
						a*sin(lineproplist[s].phi*180./PI)/c);
					L.SetValue(meshele[i].p[j],K);
					
					// second point on the side;
					x=meshnode[meshele[i].p[k]].x;
					y=meshnode[meshele[i].p[k]].y;
					r=sqrt(x*x+y*y);
					if((x==0) && (y==0)) t=0;
					else t=atan2(y,x)*180./PI;
					r*=units[LengthUnits];
					s=meshele[i].e[j];
					a=lineproplist[s].A0 + r*lineproplist[s].A1 +
					  t*lineproplist[s].A2;
					K.Set(a*cos(lineproplist[s].phi*180./PI)/c,
						a*sin(lineproplist[s].phi*180./PI)/c);
					L.SetValue(meshele[i].p[k],K);
				}

			}
		}

		if ((SDIflag==TRUE) && (sdi_iter==1)) for(i=0;i<NumEls;i++)
		for(j=0;j<3;j++){
			k=j+1; if(k==3) k=0;
			if(meshele[i].e[j]>=0)
			if(lineproplist[ meshele[i].e[j] ].BdryFormat==3)
			{			
					L.SetValue(meshele[i].p[j],0.*I);
					L.SetValue(meshele[i].p[k],0.*I);
			}
		}
	
	// "fix" diagonal entries associated with circuits that have 
	// applied current or voltage that is known a priori
	// so that solver doesn't throw a "singular" flag
	for(j=0;j<NumCircProps;j++)
		if (circproplist[j].Case<2)	L.Put(L.Get(0,0),NumNodes+j,NumNodes+j);
	
	// Apply any periodicity/antiperiodicity boundary conditions that we have
	if (NumPBCs>0)
	{
		TheView->SetDlgItemText(IDC_FRAME1,"Applying (Anti)Periodic BCs");
		TheView->m_prog1.SetPos(0);
		pctr=0;	
	}
	for(k=0;k<NumPBCs;k++)
	{
		if (pbclist[k].t==0) L.Periodicity(pbclist[k].x,pbclist[k].y);
		if (pbclist[k].t==1) L.AntiPeriodicity(pbclist[k].x,pbclist[k].y);

		// update progress bar...
		j=(k*20)/NumPBCs+1;
		if(j>pctr){
			j=pctr*5; if (j>100) j=100;
			TheView->m_prog1.SetPos(j);
			pctr++;
		}
	}

	// solve the problem;
	if (L.PBCGSolve(sdi_iter)==FALSE) return FALSE;
	if ((SDIflag==TRUE) && (sdi_iter==0))
		for (i=0;i<NumNodes+NumCircProps;i++) V_sdi[i]=L.V[i];
	if (sdi_iter==1)
		for (i=0;i<NumNodes+NumCircProps;i++)
			L.V[i]=(V_sdi[i]+L.V[i])/2.;
	
} //end of SDI loop;

	for (i=0;i<NumNodes;i++) L.b[i]=(L.V[i]*c);	// convert answer back to AMPS
	for (i=0;i<NumCircProps;i++)
		L.b[NumNodes+i]=(I*c*w*L.V[NumNodes+i]);
	// free up space allocated in this routine
	for(k=0;k<NumBlockProps;k++) free(Mu[k]);
	free(Mu);
	if(SDIflag==TRUE) free(V_sdi);
	if(NumCircProps>0){
		free(CircInt1);
		free(CircInt2);
		free(CircInt3);
	}

	return TRUE;
}

BOOL CFemmeDocCore::WriteHarmonic2D(CBigComplexLinProb &L)
{
	// write solution to disk;
		
	char c[1024];
	FILE *fp,*fz;
	int i;
	double cf;
	double unitconv[]={2.54,0.1,1.,100.,0.00254,1.e-04};
	
	// first, echo input .fem file to the .ans file;
	sprintf(c,"%s.fem",PathName);
	if((fz=fopen(c,"rt"))==NULL){
		fprintf(stderr,"Couldn't open %s.fem\n",PathName);
		return FALSE;
	}

	sprintf(c,"%s.ans",PathName);
	if((fp=fopen(c,"wt"))==NULL){
		fprintf(stderr,"Couldn't write to %s.ans\n",PathName);
		return FALSE;
	}

	while(fgets(c,1024,fz)!=NULL) fputs(c,fp);
	fclose(fz);

	// then print out node, line, and element information
	fprintf(fp,"[Solution]\n");
	cf=unitconv[LengthUnits];
	fprintf(fp,"%i\n",NumNodes);
	for(i=0;i<NumNodes;i++)
		fprintf(fp,"%.17g	%.17g	%.17g	%.17g\n",meshnode[i].x/cf,
			meshnode[i].y/cf,L.b[i].re,L.b[i].im);
	fprintf(fp,"%i\n",NumEls);
	for(i=0;i<NumEls;i++)
		fprintf(fp,"%i	%i	%i	%i\n",meshele[i].p[0],meshele[i].p[1],
			meshele[i].p[2],meshele[i].lbl);

	// last, print out circuit info
	fprintf(fp,"%i\n",NumCircProps);
	for(i=0;i<NumCircProps;i++){
		if (circproplist[i].Case==0)
			fprintf(fp,"0	%.17g	%.17g\n",circproplist[i].dV.Re(),
								      circproplist[i].dV.Im());
		if (circproplist[i].Case==1)
			fprintf(fp,"1	%.17g	%.17g\n",circproplist[i].J.Re(),
									  circproplist[i].J.Im());

		if (circproplist[i].Case==2)
			fprintf(fp,"0	%.17g	%.17g\n",L.b[NumNodes+i].Re(),
									  L.b[NumNodes+i].Im());
	}

	fclose(fp);
	return TRUE;
}